# [[[cog import cog; cog.outl('"""\n%s\n"""' % file('README.rst').read()) ]]]
from typing import Dict, List, Set, Tuple

"""
S-expression parser for Python
==============================

`sexpdata` is a simple S-expression parser/serializer.  It has
simple `load` and `dump` functions like `pickle`, `json` or `PyYAML`
module.

>>> from sexpdata import loads, dumps
>>> loads('("a" "b")')
['a', 'b']
>>> print(dumps(['a', 'b']))
("a" "b")


You can install `sexpdata` from PyPI_::

  pip install sexpdata


Links:

* `Documentation (at Read the Docs) <http://sexpdata.readthedocs.org/>`_
* `Repository (at GitHub) <https://github.com/tkf/sexpdata>`_
* `Issue tracker (at GitHub) <https://github.com/tkf/sexpdata/issues>`_
* `PyPI <http://pypi.python.org/pypi/sexpdata>`_
* `Travis CI <https://travis-ci.org/#!/tkf/sexpdata>`_


License
-------

`sexpdata` is licensed under the terms of the BSD 2-Clause License.
See the source code for more information.

"""
# [[[end]]]

# Copyright (c) 2012 Takafumi Arakaki
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__version__ = "1.0.4"
__author__ = "Joshua D. Boyd, Takafumi Arakaki"
__license__ = "BSD License"
__all__ = [
    # API functions:
    "load",
    "loads",
    "dump",
    "dumps",
    "parse",
    # Utility functions:
    "car",
    "cdr",
    # S-expression classes:
    "Symbol",
    "String",
    "Quoted",
    "Brackets",
    "Parens",
    # Position tracking:
    "Position",
    # Exception classes:
    "SExpError",
    "ExpectClosingBracket",
    "ExpectNothing",
    "ExpectSExp",
    "UnterminatedString",
    "InvalidEscape",
]

import re
from collections import namedtuple
from collections.abc import Iterable, Mapping, Sequence
from functools import singledispatch
from itertools import chain
from string import whitespace
from typing import Any, Optional

# Position Tracking


class Position:
    """Represents a position in the source text with line and column numbers."""

    def __init__(self, line: int = 1, column: int = 1, offset: int = 0) -> None:
        self.line = line
        self.column = column
        self.offset = offset

    def __repr__(self) -> str:
        return f"Position(line={self.line}, column={self.column})"

    def __str__(self) -> str:
        return f"line {self.line}, column {self.column}"


# Interface


def load(filelike, **kwds):
    """
    Load object from S-expression stored in `filelike`.

    :arg  filelike: A text stream object.

    See :func:`loads` for valid keyword arguments.

    >>> import io
    >>> fp = io.StringIO()
    >>> sexp = [Symbol('a'), Symbol('b')]   # let's dump and load this object
    >>> dump(sexp, fp)
    >>> _ = fp.seek(0)
    >>> load(fp) == sexp
    True

    """
    return loads(filelike.read(), **kwds)


def loads(string: str, **kwds):
    """
    Load object from S-expression `string`.

    :arg        string: String containing an S-expression.
    :type          nil: str or None
    :keyword       nil: A symbol interpreted as an empty list.
                        Default is ``'nil'``.
    :type         true: str or None
    :keyword      true: A symbol interpreted as True.
                        Default is ``'t'``.
    :type        false: str or None
    :keyword     false: A symbol interpreted as False.
                        Default is ``None``.
    :type     line_comment: str
    :keyword  line_comment: Beginning of line comment.
                            Default is ``';'``.

    >>> loads("(a b)")
    [Symbol('a'), Symbol('b')]
    >>> loads("a")
    Symbol('a')
    >>> loads("(a 'b)")
    [Symbol('a'), Quoted(Symbol('b'))]
    >>> loads("(a '(b))")
    [Symbol('a'), Quoted([Symbol('b')])]
    >>> loads('''
    ... ;; This is a line comment.
    ... ("a" "b")  ; this is also a comment.
    ... ''')
    ['a', 'b']
    >>> loads('''
    ... # This is a line comment.
    ... ("a" "b")  # this is also a comment.
    ... ''', line_comment='#')
    ['a', 'b']

    ``nil`` is converted to an empty list by default.  You can use
    keyword argument `nil` to change what symbol must be interpreted
    as nil:

    >>> loads("nil")
    []
    >>> loads("null", nil='null')
    []
    >>> loads("nil", nil=None)
    Symbol('nil')

    ``t`` is converted to True by default.  You can use keyword
    argument `true` to change what symbol must be converted to True.:

    >>> loads("t")
    True
    >>> loads("#t", true='#t')
    True
    >>> loads("t", true=None)
    Symbol('t')

    No symbol is converted to False by default.  You can use keyword
    argument `false` to convert a symbol to False.

    >>> loads("#f")
    Symbol('#f')
    >>> loads("#f", false='#f')
    False
    >>> loads("nil", false='nil', nil=None)
    False

    """
    obj = parse(string, **kwds)
    assert len(obj) == 1  # FIXME: raise an appropriate error
    return obj[0]


def dump(obj, filelike, **kwds) -> None:
    """
    Write `obj` as an S-expression into given stream `filelike`.

    :arg       obj: A Python object.
    :arg  filelike: A text stream object.

    See :func:`dumps` for valid keyword arguments.

    >>> import io
    >>> fp = io.StringIO()
    >>> dump(('a', 'b'), fp, str_as='symbol')
    >>> print(fp.getvalue())
    (a b)

    """
    filelike.write(dumps(obj, **kwds))


def dumps(obj, **kwds) -> str:
    """
    Convert python object into an S-expression.

    :arg           obj: A Python object.
    :type       str_as: ``'symbol'`` or ``'string'``
    :keyword    str_as: How string should be interpreted.
                        Default is ``'string'``.
    :type     tuple_as: ``'list'`` or ``'array'``
    :keyword  tuple_as: How tuple should be interpreted.
                        Default is ``'list'``.
    :type      true_as: str
    :keyword   true_as: How True should be interpreted.
                        Default is ``'t'``
    :type     false_as: str
    :keyword  false_as: How False should be interpreted.
                        Default is ``'()'``
    :type      none_as: str
    :keyword   none_as: How None should be interpreted.
                        Default is ``'()'``
    :type     pretty_print: bool
    :keyword  pretty_print: Format output as a tree.
                           Default is ``False``
    :type     indent_as: str
    :keyword  indent_as: String to use for each level of tree indentation.
                         Default is ``'  '``

    Basic usage:

    >>> print(dumps(['a', 'b']))
    ("a" "b")
    >>> print(dumps(['a', 'b'], str_as='symbol'))
    (a b)
    >>> print(dumps(dict(a=1)))
    (:a 1)
    >>> ProperTuple = namedtuple('ProperTuple', 'k')
    >>> print(dumps(ProperTuple('v')))
    (:k "v")
    >>> print(dumps([None, True, False, ()]))
    (() t () ())
    >>> print(dumps([None, True, False, ()],
    ...             none_as='null', true_as='#t', false_as='#f'))
    (null #t #f ())
    >>> print(dumps(('a', 'b')))
    ("a" "b")
    >>> print(dumps(('a', 'b'), tuple_as='array'))
    ["a" "b"]

    More verbose usage:

    >>> print(dumps([Symbol('a'), Symbol('b')]))
    (a b)
    >>> print(dumps(Symbol('a')))
    a
    >>> print(dumps([Symbol('a'), Quoted(Symbol('b'))]))
    (a 'b)
    >>> print(dumps([Symbol('a'), Quoted([Symbol('b')])]))
    (a '(b))

    """
    return str(tosexp(obj, **kwds))


def car(obj):
    """
    Alias of ``obj[0]``.

    >>> car(loads('(a . b)'))
    Symbol('a')
    >>> car(loads('(a b)'))
    Symbol('a')

    """
    return obj[0]


def cdr(obj):
    """
    `cdr`-like function.

    >>> cdr(loads('(a . b)'))
    Symbol('b')
    >>> cdr(loads('(a b)'))
    [Symbol('b')]
    >>> cdr(loads('(a . (b))'))
    [Symbol('b')]
    >>> cdr(loads('(a)'))
    []
    >>> cdr(loads('(a . nil)'))
    []

    """
    # This is very lazy implementation.  Probably the best way to do
    # it is to define `Cons` S-expression class.
    if len(obj) > 2:
        if obj[1] == Symbol("."):
            return obj[2]
    return obj[1:]


# Core


@singledispatch
def tosexp(obj, **kwds):
    """
    Convert an object to an S-expression (`dumps` is just calling this).

    See this table for comparison of lispy languages, to support them
    as much as possible:
    `Lisp: Common Lisp, Scheme/Racket, Clojure, Emacs Lisp - Hyperpolyglot
    <http://hyperpolyglot.org/lisp>`_

    Most classes can be supported by tosexp() by adding a __to_lisp_as__ method
    that returns a restructuring of an instance. The method can use builtin
    types, sexpdata hinting classes, and instances of classes that have
    tosexp() support.

    Methods that require customizing the recursion or output string of tosexp()
    should be registered with @sexpdata.tosexp.register(). Also the default
    handlers can be overridden by re-registration.

    Define tosexp() for a simple immutable Cons class. The dot is formatted
    rather than doing a 3-tuple w/Symbol('.') hack.

    >>> import sexpdata
    >>> class Cons(namedtuple('Cons', 'car cdr')):
    ...     pass
    >>> @sexpdata.tosexp.register(Cons)
    ... def _(obj, **kwds):
    ...     return '({0} . {1})'.format(sexpdata.tosexp(obj.car, **kwds),
    ...                                 sexpdata.tosexp(obj.cdr, **kwds))
    ...
    >>> dumps(Cons(True, False))
    '(t . ())'

    A simple alist using Cons:

    >>> dumps(map(Cons, 'abcde', range(5)), str_as='symbol')
    '((a . 0) (b . 1) (c . 2) (d . 3) (e . 4))'

    Overriding the float handler for application-wide formatting:

    >>> @sexpdata.tosexp.register(float)
    ... def _(obj, **kwds):
    ...     return '{0:.3}'.format(obj)
    ...
    >>> import math
    >>> tuple(round(math.pi, n) for n in range(5)) # doctest: +SKIP
    (3.0, 3.1, 3.14, 3.142, 3.1416)
    >>> dumps(round(math.pi, n) for n in range(5))
    '(3.0 3.1 3.14 3.14 3.14)'
    """
    if hasattr(obj, "__to_lisp_as__"):
        return tosexp(obj.__to_lisp_as__(), **kwds)
    else:
        raise TypeError(
            "Object of type '{0}' cannot be converted by `tosexp`. "
            "It's value is '{1!r}'".format(type(obj), obj)
        )


@tosexp.register(Iterable)
@tosexp.register(Mapping)
def _(obj, **kwds) -> str:
    return tosexp(Parens(obj), **kwds)


@tosexp.register(tuple)
def _(obj: Tuple[Any, ...], tuple_as: str = "list", **kwds) -> str:
    kwds["tuple_as"] = tuple_as
    if hasattr(obj, "__to_lisp_as__"):
        return tosexp(obj.__to_lisp_as__(), **kwds)
    elif hasattr(obj, "_asdict"):
        return tosexp(Parens(obj._asdict()), **kwds)
    elif tuple_as == "list":
        return tosexp(Parens(obj), **kwds)
    elif tuple_as == "array":
        return tosexp(Brackets(obj), **kwds)
    else:
        raise ValueError("tuple_as={0!r} is not valid".format(tuple_as))


@tosexp.register(str)
def _(obj: str, str_as: str = "string", **kwds) -> str:
    kwds["str_as"] = str_as
    if str_as == "symbol":
        return obj
    elif str_as == "string":
        return tosexp(String(obj))
    else:
        raise ValueError("str_as={0!r} is not valid".format(str_as))


@tosexp.register(type(None))
def _(obj, none_as: str = "()", **kwds) -> str:
    return none_as


@tosexp.register(bool)
def _(obj: bool, false_as: str = "()", true_as: str = "t", **kwds) -> str:
    return true_as if obj else false_as


@tosexp.register(float)
@tosexp.register(int)
def _(obj, **kwds) -> str:
    return str(obj)


class String:
    def __init__(self, object, position=None) -> None:
        self._s = str(object)
        self.position = position

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and str.__eq__(self._s, other._s)

    def __ne__(self, other) -> bool:
        return not self == other

    def __hash__(self) -> int:
        """
        >>> D = {'a': 1, String('a'): 2, Symbol('a'): 3}
        >>> len(D)
        3
        """
        return str.__hash__(self._s)

    _lisp_quoted_specials = [  # from Pymacs
        ("\\", "\\\\"),  # must come first to avoid doubly quoting "\"
        ('"', '\\"'),
        ("\b", "\\b"),
        ("\f", "\\f"),
        ("\n", "\\n"),
        ("\r", "\\r"),
        ("\t", "\\t"),
    ]

    _lisp_quoted_to_raw = dict((q, r) for (r, q) in _lisp_quoted_specials)

    def __repr__(self) -> str:
        return "{0}({1})".format(self.__class__.__name__, str.__repr__(self._s))

    def __str__(self) -> str:
        return self._s

    @classmethod
    def quote(cls, string) -> str:
        for s, q in cls._lisp_quoted_specials:
            if isinstance(string, str):
                string = string.replace(s, q)
            else:
                string = string._s.replace(s, q)
        return string

    @classmethod
    def unquote(cls, string) -> str:
        result = cls._lisp_quoted_to_raw.get(string, string)
        return result if result is not None else string

    def value(self) -> str:
        return str(self._s)


@tosexp.register(String)
def _(obj: String, **kwds) -> str:
    return '"' + String.quote(obj) + '"'


class Symbol(String):
    _lisp_quoted_specials = [
        ("\\", "\\\\"),  # must come first to avoid doubly quoting "\"
        ("'", r"\'"),
        ("`", r"\`"),
        ('"', r"\""),
        ("(", r"\("),
        (")", r"\)"),
        ("[", r"\["),
        ("]", r"\]"),
        (" ", r"\ "),
        (",", r"\,"),
        ("?", r"\?"),
        (";", r"\;"),
        ("#", r"\#"),
    ]

    _lisp_quoted_to_raw = dict((q, r) for (r, q) in _lisp_quoted_specials)


@tosexp.register(Symbol)
def _(obj: Symbol, **kwds) -> str:
    return Symbol.quote(obj)


class Quoted(namedtuple("Quoted", "x")):
    def __repr__(self) -> str:
        return "{0.__class__.__name__}({0.x!r})".format(self)


@tosexp.register(Quoted)
def _(obj: Quoted, **kwds) -> str:
    return "'" + tosexp(obj.x, **kwds)


class Delimiters(namedtuple("Delimiters", "I")):
    opener: str = ""
    closer: str = ""

    def __new__(cls, *args):
        if not args:
            raise ValueError("Expected an Iterable/Mapping argument or *args")
        x = args[0] if len(args) == 1 else args

        if isinstance(x, Mapping):
            plist_pairs = ((Symbol(":" + k), v) for k, v in x.items())
            return tuple.__new__(cls, (tuple(chain.from_iterable(plist_pairs)),))
        elif isinstance(x, (str, bytes)) or not isinstance(x, Iterable):
            return tuple.__new__(cls, ((x,),))  # unary *args
        elif isinstance(x, Sequence):
            return tuple.__new__(cls, (x,))
        else:  # isinstance(x, Iterable)
            return tuple.__new__(cls, (tuple(x),))

    @staticmethod
    def from_opener(opener: str, val):
        cls_map = dict((cls.opener, cls) for cls in Delimiters.__subclasses__())
        if opener in cls_map.keys():
            return cls_map[opener](val)
        else:
            raise TypeError

    @staticmethod
    def get_brackets() -> Dict[str, str]:
        return {cls.opener: cls.closer for cls in Delimiters.__subclasses__()}


@tosexp.register(Delimiters)
def _(self: Delimiters, **kwds) -> str:
    # Don't break up expressions produced by certain overloads of tosexp
    dont_break = all(
        tosexp.dispatch(type(x)) not in DONT_BREAK_OVERLOADS for x in self.I
    )

    if "pretty_print" in kwds and kwds["pretty_print"] and not dont_break:
        expr_separator = "\n"
        exprs_indent = kwds["indent_as"] if "indent_as" in kwds else "  "
        exprs_separator = "\n"
    else:
        expr_separator = " "
        exprs_indent = ""
        exprs_separator = ""

    exprs = expr_separator.join(tosexp(x, **kwds) for x in self.I)
    indented_exprs = "".join(exprs_indent + line for line in exprs.splitlines(True))

    return (
        self.__class__.opener
        + exprs_separator
        + indented_exprs
        + exprs_separator
        + self.__class__.closer
    )


DONT_BREAK_OVERLOADS = [
    tosexp.dispatch(c) for c in (object, Iterable, Mapping, tuple, Delimiters)
]


class Brackets(Delimiters):
    """
    Outputs an Iterable or Mapping with square brackets.

    Selectively make a container an array:

    >>> dumps(Brackets(list(range(5))))
    '[0 1 2 3 4]'

    >>> dumps(Brackets(dict(a=1)))
    '[:a 1]'
    """

    opener, closer = "[", "]"


class Parens(Delimiters):
    """
    Outputs an Iterable or Mapping with parentheses.

    By default Iterables and Mappings output with parentheses.

    >>> dumps(range(5))
    '(0 1 2 3 4)'
    >>> dumps(dict(a=1))
    '(:a 1)'

    Selectively override the tuple_as='array' default parameter:

    >>> dumps((0, Parens((1, 2, 3)), 4), tuple_as='array')
    '[0 (1 2 3) 4]'
    """

    opener, closer = "(", ")"


def bracket(val, bra: str):
    if bra == "(":
        return val
    else:
        return Delimiters.from_opener(bra, val)


class SExpError(Exception):
    """Base class for S-expression parsing errors with position information."""

    def __init__(self, message: str, position=None) -> None:
        self.position = position
        if position:
            super().__init__(f"{message} at {position}")
        else:
            super().__init__(message)


class ExpectClosingBracket(SExpError):
    def __init__(self, got, expect: str, position=None) -> None:
        message = (
            "Not enough closing brackets. "
            f"Expected {expect!r} to be the last letter in the sexp. "
            f"Got: {got!r}"
        )
        super().__init__(message, position)


class ExpectNothing(SExpError):
    def __init__(self, got, position=None) -> None:
        message = (
            "Too many closing brackets. "
            f"Expected no character left in the sexp. "
            f"Got: {got!r}"
        )
        super().__init__(message, position)


class ExpectSExp(SExpError):
    def __init__(self, position=None) -> None:
        message = "No s-exp is found after an apostrophe"
        super().__init__(message, position)


class UnterminatedString(SExpError):
    def __init__(self, position=None) -> None:
        message = "Unterminated string literal"
        super().__init__(message, position)


class InvalidEscape(SExpError):
    def __init__(self, escape_char: str, position=None) -> None:
        message = f"Invalid escape sequence: \\{escape_char}"
        super().__init__(message, position)


class Parser(object):
    brackets: Dict[str, str]
    closing_brackets: Set[str]
    _atom_end_basic: Set[str]
    _atom_end_basic_or_escape_regexp: str

    def __init__(
        self,
        string: str,
        string_to=None,
        nil: str = "nil",
        true: str = "t",
        false: Optional[str] = None,
        line_comment: str = ";",
    ) -> None:
        self.string = string
        self.nil = nil
        self.true = true
        self.false = false
        self.string_to = (lambda x: x) if string_to is None else string_to
        self.line_comment = line_comment

        # Build position mapping for error reporting
        self.position_map = self._build_position_map(string)

        # Compute brackets from delimiter
        self.brackets = Delimiters.get_brackets()
        self.closing_brackets = set(self.brackets.values())
        self._atom_end_basic = (
            set(self.brackets) | set(self.closing_brackets) | set('"') | set(whitespace)
        )
        self._atom_end_basic_or_escape_regexp = "|".join(
            map(re.escape, self._atom_end_basic | set("\\"))
        )
        self.quote_or_escape_re = re.compile(r'"|\\')
        self.atom_end = set([line_comment]) | self._atom_end_basic
        self.atom_end_or_escape_re = re.compile(
            "{0}|{1}".format(
                self._atom_end_basic_or_escape_regexp, re.escape(line_comment)
            )
        )

    def _build_position_map(self, string: str) -> Dict[int, Position]:
        """Build a mapping from string offset to (line, column)"""
        positions = {}
        line, column = 1, 1
        for i, char in enumerate(string):
            positions[i] = Position(line, column, i)
            if char == "\n":
                line += 1
                column = 1
            else:
                column += 1
        positions[len(string)] = Position(line, column, len(string))
        return positions

    def get_position(self, offset: int) -> Position:
        """Get position information for a given string offset."""
        return self.position_map.get(offset, Position())

    def parse_str(self, i: int) -> Tuple[int, str]:
        string = self.string
        chars: List[str] = []
        append = chars.append
        search = self.quote_or_escape_re.search

        start_pos = self.get_position(i)
        while True:
            i += 1
            match = search(string, i)
            if match is None:
                raise UnterminatedString(start_pos)
            end = match.start()
            append(string[i:end])
            c = match.group()
            if c == '"':
                i = end + 1
                break
            elif c == "\\":
                i = end + 1
                if i >= len(string):
                    raise InvalidEscape("EOF", self.get_position(end))
                append(String.unquote(c + string[i]))
        else:
            raise ExpectClosingBracket('"', self.get_position())
        return (i, "".join(chars))

    def parse_atom(self, i: int) -> Tuple[int, Any]:
        string = self.string
        chars: List[str] = []
        append = chars.append
        search = self.atom_end_or_escape_re.search
        atom_end = self.atom_end
        start_pos = self.get_position(i)

        while True:
            match = search(string, i)
            if not match:
                append(string[i:])
                i = len(string)
                break
            end = match.start()
            append(string[i:end])
            c = match.group()
            if c in atom_end:
                i = end  # this is different from str
                break
            elif c == "\\":
                i = end + 1
                if i >= len(string):
                    raise InvalidEscape("EOF", self.get_position(end))
                next_char = string[i]
                try:
                    unquoted = Symbol.unquote(c + next_char)
                    append(unquoted)
                except KeyError:
                    raise InvalidEscape(next_char, self.get_position(i))
            i += 1

        return (i, self.atom("".join(chars), start_pos))

    def atom(self, token: str, position=None):
        if token == self.nil:
            return []
        if token == self.true:
            return True
        if token == self.false:
            return False
        try:
            return int(token)
        except ValueError:
            try:
                result = float(token)
                # Block automatic conversion to infinity or NaN
                if (
                    result in (float("inf"), float("-inf")) or result != result
                ):  # NaN check
                    raise ValueError("Invalid s-exp float")
                return result
            except ValueError:
                return Symbol(token, position)

    def parse_sexp(self, i: int) -> Tuple[int, List[Any]]:
        string = self.string
        len_string = len(self.string)
        sexp: List[Any] = []
        append = sexp.append
        bracket_stack = []  # Track opening brackets for better error reporting

        while i < len_string:
            c = string[i]
            current_pos = self.get_position(i)

            if c == '"':
                try:
                    (i, parsed_str) = self.parse_str(i)
                    append(self.string_to(parsed_str))
                except SExpError:
                    raise  # Re-raise with position already set
            elif c in whitespace:
                i += 1
                continue
            elif c in self.brackets:
                close = self.brackets[c]
                bracket_stack.append((c, close, current_pos))
                (next_i, parsed_sexp) = self.parse_sexp(i + 1)
                i = next_i
                append(bracket(parsed_sexp, c))
                try:
                    nc = string[i]
                except IndexError:
                    bracket_info = bracket_stack[-1] if bracket_stack else None
                    if bracket_info:
                        raise ExpectClosingBracket(
                            None, bracket_info[1], bracket_info[2]
                        )
                    else:
                        raise ExpectClosingBracket(None, close, current_pos)
                if nc != close:
                    raise ExpectClosingBracket(nc, close, current_pos)
                bracket_stack.pop() if bracket_stack else None
                i += 1
            elif c in self.closing_brackets:
                break
            elif c == "'":
                quote_pos = current_pos
                next_parse_start = i + 1
                (next_i, parsed_sexp) = self.parse_sexp(next_parse_start)
                i = next_i
                if not parsed_sexp:
                    raise ExpectSExp(quote_pos)
                append(Quoted(parsed_sexp[0]))
                sexp.extend(parsed_sexp[1:])
            elif c == self.line_comment:
                i = string.find("\n", i) + 1
                if i <= 0:
                    i = len_string
                    break
            else:
                (i, parsed_atom) = self.parse_atom(i)
                append(parsed_atom)
        return (i, sexp)

    def parse(self) -> List[Any]:
        try:
            (i, sexp) = self.parse_sexp(0)
            if i < len(self.string):
                raise ExpectNothing(self.string[i:], self.get_position(i))
            return sexp
        except SExpError:
            raise  # Re-raise S-expression specific errors
        except Exception as e:
            # Wrap unexpected errors with position info if possible
            raise SExpError(f"Unexpected parsing error: {e}")


def parse(string: str, **kwds) -> List[Any]:
    """
    Parse s-expression.

    >>> parse("(a b)")
    [[Symbol('a'), Symbol('b')]]
    >>> parse("a")
    [Symbol('a')]
    >>> parse("(a 'b)")
    [[Symbol('a'), Quoted(Symbol('b'))]]
    >>> parse("(a '(b))")
    [[Symbol('a'), Quoted([Symbol('b')])]]

    """
    assert isinstance(string, str)
    return Parser(string, **kwds).parse()



"""Common enumeration types for KiCad S-expressions."""

from enum import Enum


class PinElectricalType(Enum):
    """Pin electrical types for symbols."""

    INPUT = "input"
    OUTPUT = "output"
    BIDIRECTIONAL = "bidirectional"
    TRI_STATE = "tri_state"
    PASSIVE = "passive"
    FREE = "free"
    UNSPECIFIED = "unspecified"
    POWER_IN = "power_in"
    POWER_OUT = "power_out"
    OPEN_COLLECTOR = "open_collector"
    OPEN_EMITTER = "open_emitter"
    NO_CONNECT = "no_connect"


class PinGraphicStyle(Enum):
    """Pin graphical styles for symbols."""

    LINE = "line"
    INVERTED = "inverted"
    CLOCK = "clock"
    INVERTED_CLOCK = "inverted_clock"
    INPUT_LOW = "input_low"
    CLOCK_LOW = "clock_low"
    OUTPUT_LOW = "output_low"
    EDGE_CLOCK_HIGH = "edge_clock_high"
    NON_LOGIC = "non_logic"


class PadType(Enum):
    """Pad types for footprints."""

    THRU_HOLE = "thru_hole"
    SMD = "smd"
    CONNECT = "connect"
    NP_THRU_HOLE = "np_thru_hole"


class PadShape(Enum):
    """Pad shapes for footprints."""

    CIRCLE = "circle"
    RECT = "rect"
    OVAL = "oval"
    TRAPEZOID = "trapezoid"
    ROUNDRECT = "roundrect"
    CUSTOM = "custom"


class StrokeType(Enum):
    """Valid stroke line styles for graphics."""

    DASH = "dash"
    DASH_DOT = "dash_dot"
    DASH_DOT_DOT = "dash_dot_dot"
    DOT = "dot"
    DEFAULT = "default"
    SOLID = "solid"


class JustifyHorizontal(Enum):
    """Horizontal text justification."""

    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"


class JustifyVertical(Enum):
    """Vertical text justification."""

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"


class FillType(Enum):
    """Fill types for graphical objects."""

    NONE = "none"
    OUTLINE = "outline"
    BACKGROUND = "background"
    COLOR = "color"


class LabelShape(Enum):
    """Label and pin shapes for global labels, hierarchical labels, and sheet pins."""

    INPUT = "input"
    OUTPUT = "output"
    BIDIRECTIONAL = "bidirectional"
    TRI_STATE = "tri_state"
    PASSIVE = "passive"


class FootprintTextType(Enum):
    """Footprint text types."""

    REFERENCE = "reference"
    VALUE = "value"
    USER = "user"


class LayerType(Enum):
    """PCB layer types."""

    SIGNAL = "signal"
    POWER = "power"
    MIXED = "mixed"
    JUMPER = "jumper"
    USER = "user"


class ViaType(Enum):
    """Via types for PCB."""

    THROUGH = "through"
    BLIND_BURIED = "blind_buried"
    MICRO = "micro"


class ZoneConnection(Enum):
    """Zone connection types for pads."""

    INHERITED = 0
    SOLID = 1
    THERMAL_RELIEF = 2
    NONE = 3


class ZoneFillMode(Enum):
    """Zone fill modes."""

    SOLID = "solid"
    HATCHED = "hatched"


class ZoneKeepoutSetting(Enum):
    """Zone keepout settings."""

    ALLOWED = "allowed"
    NOT_ALLOWED = "not_allowed"


class HatchStyle(Enum):
    """Zone hatch display styles."""

    NONE = "none"
    EDGE = "edge"
    FULL = "full"


class SmoothingStyle(Enum):
    """Zone corner smoothing styles."""

    NONE = "none"
    CHAMFER = "chamfer"
    FILLET = "fillet"


class ClearanceType(Enum):
    """Custom pad clearance types."""

    OUTLINE = "outline"
    CONVEXHULL = "convexhull"


class SeverityLevel(Enum):
    """Design rule severity levels."""

    ERROR = "error"
    WARNING = "warning"
    IGNORE = "ignore"
    EXCLUSION = "exclusion"


class ConstraintType(Enum):
    """Design rule constraint types."""

    # Clearance Constraints
    CLEARANCE = "clearance"
    HOLE_CLEARANCE = "hole_clearance"
    EDGE_CLEARANCE = "edge_clearance"
    SILK_CLEARANCE = "silk_clearance"
    COURTYARD_CLEARANCE = "courtyard_clearance"
    HOLE_TO_HOLE = "hole_to_hole"
    PHYSICAL_CLEARANCE = "physical_clearance"
    PHYSICAL_HOLE_CLEARANCE = "physical_hole_clearance"
    CREEPAGE = "creepage"

    # Size Constraints
    TRACK_WIDTH = "track_width"
    VIA_DIAMETER = "via_diameter"
    HOLE_SIZE = "hole_size"
    ANNULAR_WIDTH = "annular_width"
    CONNECTION_WIDTH = "connection_width"

    # Thermal Constraints
    THERMAL_RELIEF_GAP = "thermal_relief_gap"
    THERMAL_SPOKE_WIDTH = "thermal_spoke_width"
    MIN_RESOLVED_SPOKES = "min_resolved_spokes"

    # Zone Constraints
    ZONE_CONNECTION = "zone_connection"

    # Disallow Constraints
    DISALLOW = "disallow"

    # Length Constraints
    LENGTH = "length"
    SKEW = "skew"

    # Differential Pair Constraints
    DIFF_PAIR_GAP = "diff_pair_gap"
    DIFF_PAIR_UNCOUPLED = "diff_pair_uncoupled"

    # Via Constraints
    VIA_DRILL = "via_drill"
    BLIND_VIA_RATIO = "blind_via_ratio"
    MICRO_VIA_DIAMETER = "micro_via_diameter"
    MICRO_VIA_DRILL = "micro_via_drill"

    # Text Constraints
    TEXT_HEIGHT = "text_height"
    TEXT_THICKNESS = "text_thickness"

    # Assertion Constraints
    ASSERTION = "assertion"



"""KiCadFiles - A comprehensive Python library for parsing and manipulating KiCad file formats.

This library provides a complete set of dataclasses representing all KiCad
S-expression tokens. Each class corresponds exactly to one S-expression token
and follows a consistent structure for parsing and serialization.

The classes are organized into logical modules based on dependencies:
- base_element: Base KiCadObject class with simplified strictness modes
- enums: Common enumeration types for type safety
- base_types: Fundamental types with no dependencies (37 classes)
- text_and_documents: Text and document elements (35 classes)
- pad_and_drill: Pad and drill elements (17 classes)
- primitive_graphics: Basic graphics primitives (8 classes)
- advanced_graphics: Complex graphics objects (20 classes)
- symbol_library: Symbol management (15 classes)
- footprint_library: Footprint management (12 classes)
- zone_system: Zone and copper filling (28 classes)
- board_layout: PCB board design (15 classes)
- schematic_system: Schematic drawing (17 classes)
- design_rules: Design rule check definitions (10 classes)
- project_settings: JSON project settings (14 classes)

Total: 241 classes representing all KiCad S-expression and JSON tokens.

Key Features:
- Simplified parser strictness modes: STRICT, FAILSAFE, SILENT
- Comprehensive type safety with Python type hints
- Round-trip parsing: parse KiCad files and convert back to S-expressions
- Zero external dependencies
- Extensive test coverage

Usage:
    from kicadfiles import At, Layer, Footprint, ParseStrictness

    # Create objects using dataclass syntax
    position = At(x=10.0, y=20.0, angle=90.0)
    layer = Layer(name="F.Cu")

    # Parse from S-expression with different strictness levels
    at_obj = At.from_sexpr("(at 10.0 20.0 90.0)", ParseStrictness.STRICT)

    # Convert back to S-expression
    sexpr_str = at_obj.to_sexpr_str()
"""

# Version information
from .__version__ import __version__, __version_info__  # noqa: F401

# Advanced graphics
from .advanced_graphics import (
    Dimension,
    Format,
    FpArc,
    FpCircle,
    FpCurve,
    FpLine,
    FpPoly,
    FpRect,
    FpText,
    FpTextBox,
    GrArc,
    GrBbox,
    GrCircle,
    GrCurve,
    GrLine,
    GrPoly,
    GrRect,
    GrText,
    GrTextBox,
)

# Base element
from .base_element import (
    KiCadFloat,
    KiCadInt,
    KiCadObject,
    KiCadStr,
    OptionalFlag,
    OptionalSimpleFlag,
    ParseStrictness,
)

# Base types
from .base_types import (
    Anchor,
    At,
    AtXY,
    BoardLayers,
    Center,
    Color,
    Effects,
    End,
    Fill,
    Font,
    Justify,
    Layer,
    LayerDefinition,
    Layers,
    Offset,
    Pos,
    Property,
    Pts,
    Size,
    Start,
    Stroke,
    Text,
    Type,
    Uuid,
    Xy,
    Xyz,
)

# Board layout
from .board_layout import (
    BoardArc,
    KicadPcb,
    Nets,
    PcbPlotParams,
    PrivateLayers,
    Segment,
    Setup,
    Stackup,
    StackupLayer,
    Tenting,
    Tracks,
    Via,
    Vias,
)

# Design rules
from .design_rules import (
    DesignRule,
    DesignRuleConstraint,
    DesignRuleSeverity,
    KiCadDesignRules,
)

# Enums
from .enums import (
    ClearanceType,
    FillType,
    FootprintTextType,
    HatchStyle,
    JustifyHorizontal,
    JustifyVertical,
    LabelShape,
    LayerType,
    PadShape,
    PadType,
    PinElectricalType,
    PinGraphicStyle,
    SmoothingStyle,
    StrokeType,
    ViaType,
    ZoneConnection,
    ZoneFillMode,
    ZoneKeepoutSetting,
)

# Footprint library
from .footprint_library import (
    Attr,
    Footprint,
    Footprints,
    Model,
    NetTiePadGroups,
)

# Library tables
from .library_tables import (
    FpLibTable,
    LibraryEntry,
    SymLibTable,
)

# Pad and drill elements
from .pad_and_drill import (
    Chamfer,
    Drill,
    Net,
    Options,
    Pad,
    Pads,
    Primitives,
    Shape,
    ZoneConnect,
)

# Primitive graphics
from .primitive_graphics import (
    Arc,
    Bezier,
    Circle,
    Line,
    Polygon,
    Polyline,
    Rect,
    Rectangle,
)

# Project settings
from .project_settings import (
    BoardDefaults,
    BoardSettings,
    CvpcbSettings,
    DesignSettings,
    ERCSettings,
    IPC2581Settings,
    KicadProject,
    LibrarySettings,
    NetClass,
    NetSettings,
    PcbnewSettings,
    ProjectMeta,
    SchematicBOMSettings,
    SchematicSettings,
)

# Schematic system
from .schematic_system import (
    Bus,
    BusEntry,
    GlobalLabel,
    HierarchicalLabel,
    Junction,
    KicadSch,
    Label,
    NetclassFlag,
    NoConnect,
    Project,
    RuleArea,
    Sheet,
    SheetInstance,
    SheetInstances,
    Table,
    Wire,
)

# Symbol library
from .symbol_library import (
    Instances,
    KicadSymbolLib,
    LibSymbols,
    Number,
    Pin,
    PinName,
    PinNames,
    PinNumbers,
    Pintype,
    Symbol,
)

# Templates (convenience helpers)
from .templates import (
    KiCadTemplates,
)

# Text and document elements
from .text_and_documents import (
    Bitmap,
    Comment,
    Data,
    Group,
    Image,
    KicadWks,
    Members,
    Paper,
    Pngdata,
    Tbtext,
    Textsize,
    TitleBlock,
    WksLine,
    WksRect,
    WksSetup,
    WksTbText,
    WksTextsize,
)

# Zone system
from .zone_system import (
    ConnectPads,
    Copperpour,
    FilledPolygon,
    FilledSegments,
    FillSegments,
    Hatch,
    HatchOrientation,
    Keepout,
    Mode,
    Smoothing,
    Zone,
    ZoneFill,
)

# Grid helper constants
GRID = 1.27  # KiCad schematic default grid in mm (50 mil)

__all__ = [
    # Grid helpers
    "GRID",
    # Base
    "KiCadFloat",
    "KiCadInt",
    "KiCadObject",
    "KiCadStr",
    "ParseStrictness",
    "OptionalFlag",
    "OptionalSimpleFlag",
    # Enums
    "ClearanceType",
    "FillType",
    "FootprintTextType",
    "HatchStyle",
    "JustifyHorizontal",
    "JustifyVertical",
    "LabelShape",
    "LayerType",
    "PadShape",
    "PadType",
    "PinElectricalType",
    "PinGraphicStyle",
    "SmoothingStyle",
    "StrokeType",
    "ViaType",
    "ZoneConnection",
    "ZoneFillMode",
    "ZoneKeepoutSetting",
    # Base types
    "Anchor",
    "At",
    "AtXY",
    "BoardLayers",
    "Center",
    "Color",
    "Effects",
    "End",
    "Fill",
    "Font",
    "Justify",
    "Layer",
    "LayerDefinition",
    "Layers",
    "Offset",
    "Pos",
    "Property",
    "Pts",
    "Size",
    "Start",
    "Stroke",
    "Text",
    "Type",
    "Uuid",
    "Xy",
    "Xyz",
    # Design rules
    "DesignRule",
    "DesignRuleConstraint",
    "DesignRuleSeverity",
    "KiCadDesignRules",
    # Text and documents
    "Bitmap",
    "Comment",
    "Data",
    "Group",
    "Image",
    "KicadWks",
    "Members",
    "Paper",
    "Pngdata",
    "Tbtext",
    "Textsize",
    "TitleBlock",
    "WksLine",
    "WksRect",
    "WksSetup",
    "WksTbText",
    "WksTextsize",
    # Pad and drill
    "Chamfer",
    "Drill",
    "Net",
    "Options",
    "Pad",
    "Pads",
    "Primitives",
    "Shape",
    "ZoneConnect",
    # Primitive graphics
    "Arc",
    "Bezier",
    "Circle",
    "Line",
    "Polygon",
    "Polyline",
    "Rect",
    "Rectangle",
    # Advanced graphics
    "Dimension",
    "Format",
    "FpArc",
    "FpCircle",
    "FpCurve",
    "FpLine",
    "FpPoly",
    "FpRect",
    "FpText",
    "FpTextBox",
    "GrArc",
    "GrBbox",
    "GrCircle",
    "GrCurve",
    "GrLine",
    "GrPoly",
    "GrRect",
    "GrText",
    "GrTextBox",
    # Symbol library
    "Instances",
    "KicadSymbolLib",
    "LibSymbols",
    "PinName",
    "Number",
    "Pin",
    "PinNames",
    "PinNumbers",
    "Pintype",
    "Symbol",
    # Footprint library
    "Attr",
    "Footprint",
    "Footprints",
    "Model",
    "NetTiePadGroups",
    # Library tables
    "FpLibTable",
    "LibraryEntry",
    "SymLibTable",
    # Project settings
    "BoardDefaults",
    "BoardSettings",
    "CvpcbSettings",
    "DesignSettings",
    "ERCSettings",
    "IPC2581Settings",
    "KicadProject",
    "LibrarySettings",
    "NetClass",
    "NetSettings",
    "PcbnewSettings",
    "ProjectMeta",
    "SchematicBOMSettings",
    "SchematicSettings",
    # Zone system
    "ConnectPads",
    "Copperpour",
    "FillSegments",
    "FilledPolygon",
    "FilledSegments",
    "Hatch",
    "HatchOrientation",
    "Keepout",
    "Mode",
    "Smoothing",
    "Zone",
    "ZoneFill",
    # Board layout
    "BoardArc",
    "KicadPcb",
    "Nets",
    "PcbPlotParams",
    "PrivateLayers",
    "Segment",
    "Setup",
    "Stackup",
    "StackupLayer",
    "Tenting",
    "Tracks",
    "Via",
    "Vias",
    # Schematic system
    "Bus",
    "BusEntry",
    "GlobalLabel",
    "HierarchicalLabel",
    "Junction",
    "KicadSch",
    "Label",
    "NetclassFlag",
    "NoConnect",
    "Project",
    "RuleArea",
    "Sheet",
    "SheetInstance",
    "SheetInstances",
    "Table",
    "Wire",
    # Templates
    "KiCadTemplates",
]



"""Library table classes for KiCad library management.

This module provides dataclasses for parsing and managing KiCad library tables:
- FpLibTable: Footprint library table (fp-lib-table)
- SymLibTable: Symbol library table (sym-lib-table)

Both table types share the same structure with library entries containing:
- name: Library identifier
- type: Library type (typically "KiCad")
- uri: Path to library
- options: Additional options
- descr: Human readable description
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from .base_element import KiCadObject, ParseStrictness


@dataclass
class LibraryEntry(KiCadObject):
    """A single library entry in a library table.

    Represents a (lib ...) entry with name, type, uri, options, and description.

    Args:
        name: Library name/identifier
        type: Library type (e.g., 'KiCad')
        uri: Path to library
        options: Additional options
        descr: Human readable description
    """

    __token_name__ = "lib"

    name: str = field(
        default="",
        metadata={"description": "Library name/identifier"},
    )
    type: str = field(
        default="", metadata={"description": "Library type (e.g., 'KiCad')"}
    )
    uri: str = field(default="", metadata={"description": "Path to library"})
    options: str = field(default="", metadata={"description": "Additional options"})
    descr: str = field(
        default="", metadata={"description": "Human readable description"}
    )


@dataclass
class FpLibTable(KiCadObject):
    """Footprint library table (fp-lib-table file).

    Contains version information and a list of footprint library entries.

    Args:
        version: Table format version
        libraries: List of library entries
    """

    __token_name__ = "fp_lib_table"

    version: int = field(default=7, metadata={"description": "Table format version"})
    libraries: List[LibraryEntry] = field(
        default_factory=list, metadata={"description": "List of library entries"}
    )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        strictness: ParseStrictness = ParseStrictness.STRICT,
        encoding: str = "utf-8",
    ) -> "FpLibTable":
        """Parse from fp-lib-table file - convenience method for library table operations."""
        if not file_path.endswith("fp-lib-table"):
            raise ValueError("Unsupported file extension. Expected: fp-lib-table")
        with open(file_path, "r", encoding=encoding) as f:
            content = f.read()
        return cls.from_str(content, strictness)

    def save_to_file(self, file_path: str, encoding: str = "utf-8") -> None:
        """Save to fp-lib-table file format.

        Args:
            file_path: Path to write the fp-lib-table file
            encoding: File encoding (default: utf-8)
        """
        if not file_path.endswith("fp-lib-table"):
            raise ValueError("Unsupported file extension. Expected: fp-lib-table")
        content = self.to_sexpr_str()
        with open(file_path, "w", encoding=encoding) as f:
            f.write(content)


@dataclass
class SymLibTable(KiCadObject):
    """Symbol library table (sym-lib-table file).

    Contains version information and a list of symbol library entries.

    Args:
        version: Table format version
        libraries: List of library entries
    """

    __token_name__ = "sym_lib_table"

    version: int = field(default=7, metadata={"description": "Table format version"})
    libraries: List[LibraryEntry] = field(
        default_factory=list, metadata={"description": "List of library entries"}
    )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        strictness: ParseStrictness = ParseStrictness.STRICT,
        encoding: str = "utf-8",
    ) -> "SymLibTable":
        """Parse from sym-lib-table file - convenience method for library table operations."""
        if not file_path.endswith("sym-lib-table"):
            raise ValueError("Unsupported file extension. Expected: sym-lib-table")
        with open(file_path, "r", encoding=encoding) as f:
            content = f.read()
        return cls.from_str(content, strictness)

    def save_to_file(self, file_path: str, encoding: str = "utf-8") -> None:
        """Save to sym-lib-table file format.

        Args:
            file_path: Path to write the sym-lib-table file
            encoding: File encoding (default: utf-8)
        """
        if not file_path.endswith("sym-lib-table"):
            raise ValueError("Unsupported file extension. Expected: sym-lib-table")
        content = self.to_sexpr_str()
        with open(file_path, "w", encoding=encoding) as f:
            f.write(content)



"""Text and document related elements for KiCad S-expressions."""

from dataclasses import dataclass, field
from typing import Any, List, Optional

from .base_element import (
    KiCadFloat,
    KiCadInt,
    KiCadObject,
    KiCadStr,
    OptionalFlag,
    ParseStrictness,
)
from .base_types import (
    AtXY,
    End,
    Font,
    Justify,
    Pos,
    Size,
    Start,
    Uuid,
)


@dataclass
class Comment(KiCadObject):
    """Comment definition token.

    The 'comment' token defines document comments in the format::

        (comment N "COMMENT")

    Where N is a number from 1 to 9.

    Args:
        number: Comment number (1-9)
        text: Comment text
    """

    __token_name__ = "comment"

    number: int = field(default=1, metadata={"description": "Comment number (1-9)"})
    text: str = field(default="", metadata={"description": "Comment text"})


@dataclass
class Data(KiCadObject):
    """Data definition token.

    The 'data' token defines hexadecimal byte data in the format::

        (data XX1 ... XXN)

    Where XXN represents hexadecimal bytes separated by spaces, with a maximum of 32 bytes per data token.

    Args:
        hex_bytes: Hexadecimal byte values (up to 32 bytes)
    """

    __token_name__ = "data"

    hex_bytes: List[str] = field(
        default_factory=list,
        metadata={"description": "Hexadecimal byte values (up to 32 bytes)"},
    )


@dataclass
class Paper(KiCadObject):
    """Paper settings definition token.

    The 'paper' token defines paper size and orientation in the format::

        (paper PAPER_SIZE | WIDTH HEIGHT [portrait])

    Where PAPER_SIZE can be: A0, A1, A2, A3, A4, A5, A, B, C, D, E.

    Args:
        size: Standard paper size (optional)
        width: Custom paper width (optional)
        height: Custom paper height (optional)
        portrait: Whether paper is in portrait mode (optional)
    """

    __token_name__ = "paper"

    size: Optional[str] = field(
        default=None, metadata={"description": "Standard paper size", "required": False}
    )
    width: Optional[float] = field(
        default=None, metadata={"description": "Custom paper width", "required": False}
    )
    height: Optional[float] = field(
        default=None, metadata={"description": "Custom paper height", "required": False}
    )
    portrait: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("portrait"),
        metadata={
            "description": "Whether paper is in portrait mode",
            "required": False,
        },
    )


@dataclass
class TitleBlock(KiCadObject):
    """Title block definition token.

    The 'title_block' token defines the document title block in the format::

        (title_block
            (title "TITLE")
            (date "DATE")
            (rev "REVISION")
            (company "COMPANY_NAME")
            (comment N "COMMENT")
        )

    Args:
        title: Document title (optional)
        date: Document date (optional)
        rev: Document revision (optional)
        company: Company name (optional)
        comments: List of comments (optional)
    """

    __token_name__ = "title_block"

    title: KiCadStr = field(
        default_factory=lambda: KiCadStr("title", "", required=False),
        metadata={"description": "Document title", "required": False},
    )
    date: KiCadStr = field(
        default_factory=lambda: KiCadStr("date", "", required=False),
        metadata={"description": "Document date", "required": False},
    )
    rev: KiCadStr = field(
        default_factory=lambda: KiCadStr("rev", "", required=False),
        metadata={"description": "Document revision", "required": False},
    )
    company: KiCadStr = field(
        default_factory=lambda: KiCadStr("company", "", required=False),
        metadata={"description": "Company name", "required": False},
    )
    comments: Optional[List[Comment]] = field(
        default_factory=list,
        metadata={"description": "List of comments", "required": False},
    )


@dataclass
class Tbtext(KiCadObject):
    """Title block text definition token.

    The 'tbtext' token defines text elements in the title block in the format::

        (tbtext
            "TEXT"
            (name "NAME")
            (pos X Y [CORNER])
            (font [(size WIDTH HEIGHT)] [bold] [italic])
            [(repeat COUNT)]
            [(incrx DISTANCE)]
            [(incry DISTANCE)]
            [(comment "COMMENT")]
        )

    Args:
        text: Text content
        name: Text element name
        pos: Position coordinates
        font: Font settings (optional)
        repeat: Repeat count for incremental text (optional)
        incrx: Repeat distance on X axis (optional)
        incry: Repeat distance on Y axis (optional)
        comment: Comment for the text object (optional)
    """

    __token_name__ = "tbtext"

    text: str = field(default="", metadata={"description": "Text content"})
    name: KiCadStr = field(
        default_factory=lambda: KiCadStr("name", ""),
        metadata={"description": "Text element name"},
    )
    pos: Pos = field(
        default_factory=lambda: Pos(), metadata={"description": "Position coordinates"}
    )
    font: Optional[Font] = field(
        default=None, metadata={"description": "Font settings", "required": False}
    )
    repeat: KiCadInt = field(
        default_factory=lambda: KiCadInt("repeat", 0, required=False),
        metadata={
            "description": "Repeat count for incremental text",
            "required": False,
        },
    )
    incrx: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incrx", 0.0, required=False),
        metadata={"description": "Repeat distance on X axis", "required": False},
    )
    incry: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incry", 0.0, required=False),
        metadata={"description": "Repeat distance on Y axis", "required": False},
    )
    comment: Optional[str] = field(
        default=None,
        metadata={"description": "Comment for the text object", "required": False},
    )


@dataclass
class Textsize(KiCadObject):
    """Text size definition token.

    The 'textsize' token defines text size in the format::

        (textsize WIDTH HEIGHT)

    Args:
        size: Text size (width and height)
    """

    __token_name__ = "textsize"

    size: Size = field(
        default_factory=lambda: Size(),
        metadata={"description": "Text size (width and height)"},
    )


@dataclass
class Members(KiCadObject):
    """Group members definition token.

    The 'members' token defines the members of a group in the format::

        (members UUID1 UUID2 ... UUIDN)

    Args:
        uuids: List of member UUIDs
    """

    __token_name__ = "members"

    uuids: List[str] = field(
        default_factory=list, metadata={"description": "List of member UUIDs"}
    )


@dataclass
class Group(KiCadObject):
    """Group definition token.

    The 'group' token defines a group of objects in the format::

        (group
            "NAME"
            (uuid UUID)
            (members UUID1 ... UUIDN)
        )

    Args:
        name: Group name
        uuid: Group unique identifier (optional)
        members: List of member UUIDs (optional)
    """

    __token_name__ = "group"

    name: str = field(default="", metadata={"description": "Group name"})
    uuid: Optional[Uuid] = field(
        default=None,
        metadata={"description": "Group unique identifier", "required": False},
    )
    members: Optional[Members] = field(
        default=None,
        metadata={"description": "List of member UUIDs", "required": False},
    )


@dataclass
class WksTextsize(KiCadObject):
    """Worksheet text size definition token.

    Args:
        width: Text width
        height: Text height
    """

    __token_name__ = "textsize"

    width: float = field(
        default=1.0,
        metadata={"description": "Text width"},
    )
    height: float = field(
        default=1.0,
        metadata={"description": "Text height"},
    )


@dataclass
class WksSetup(KiCadObject):
    """Worksheet setup definition token.

    Args:
        textsize: Text size (optional)
        linewidth: Line width (optional)
        textlinewidth: Text line width (optional)
        left_margin: Left margin (optional)
        right_margin: Right margin (optional)
        top_margin: Top margin (optional)
        bottom_margin: Bottom margin (optional)
    """

    __token_name__ = "setup"

    textsize: Optional[WksTextsize] = field(
        default=None,
        metadata={"description": "Text size", "required": False},
    )
    linewidth: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("linewidth", 0.0, required=False),
        metadata={"description": "Line width", "required": False},
    )
    textlinewidth: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("textlinewidth", 0.0, required=False),
        metadata={"description": "Text line width", "required": False},
    )
    left_margin: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("left_margin", 0.0, required=False),
        metadata={"description": "Left margin", "required": False},
    )
    right_margin: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("right_margin", 0.0, required=False),
        metadata={"description": "Right margin", "required": False},
    )
    top_margin: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("top_margin", 0.0, required=False),
        metadata={"description": "Top margin", "required": False},
    )
    bottom_margin: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("bottom_margin", 0.0, required=False),
        metadata={"description": "Bottom margin", "required": False},
    )


@dataclass
class WksRect(KiCadObject):
    """Worksheet rectangle definition token.

    Args:
        name: Rectangle name (optional)
        start: Start position
        end: End position
        comment: Comment (optional)
        repeat: Repeat count (optional)
        incrx: X increment (optional)
        incry: Y increment (optional)
        linewidth: Line width (optional)
    """

    __token_name__ = "rect"

    name: Optional[str] = field(
        default=None, metadata={"description": "Rectangle name", "required": False}
    )
    start: Start = field(
        default_factory=lambda: Start(), metadata={"description": "Start position"}
    )
    end: End = field(
        default_factory=lambda: End(), metadata={"description": "End position"}
    )
    comment: KiCadStr = field(
        default_factory=lambda: KiCadStr("comment", "", required=False),
        metadata={"description": "Comment", "required": False},
    )
    repeat: KiCadInt = field(
        default_factory=lambda: KiCadInt("repeat", 0, required=False),
        metadata={"description": "Repeat count", "required": False},
    )
    incrx: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incrx", 0.0, required=False),
        metadata={"description": "X increment", "required": False},
    )
    incry: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incry", 0.0, required=False),
        metadata={"description": "Y increment", "required": False},
    )
    linewidth: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("linewidth", 0.0, required=False),
        metadata={"description": "Line width", "required": False},
    )


@dataclass
class WksLine(KiCadObject):
    """Worksheet line definition token.

    Args:
        name: Line name (optional)
        start: Start position
        end: End position
        repeat: Repeat count (optional)
        incrx: X increment (optional)
        incry: Y increment (optional)
    """

    __token_name__ = "line"

    name: Optional[str] = field(
        default=None, metadata={"description": "Line name", "required": False}
    )
    start: Start = field(
        default_factory=lambda: Start(), metadata={"description": "Start position"}
    )
    end: End = field(
        default_factory=lambda: End(), metadata={"description": "End position"}
    )
    repeat: KiCadInt = field(
        default_factory=lambda: KiCadInt("repeat", 0, required=False),
        metadata={"description": "Repeat count", "required": False},
    )
    incrx: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incrx", 0.0, required=False),
        metadata={"description": "X increment", "required": False},
    )
    incry: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incry", 0.0, required=False),
        metadata={"description": "Y increment", "required": False},
    )


@dataclass
class WksTbText(KiCadObject):
    """Worksheet text block definition token.

    Args:
        text: Text content
        name: Text name (optional)
        pos: Text position
        font: Font settings (optional)
        justify: Text justification (optional)
        repeat: Repeat count (optional)
        incrx: X increment (optional)
        incry: Y increment (optional)
        comment: Comment (optional)
    """

    __token_name__ = "tbtext"

    text: str = field(default="", metadata={"description": "Text content"})
    name: KiCadStr = field(
        default_factory=lambda: KiCadStr("name", "", required=False),
        metadata={"description": "Text name", "required": False},
    )
    pos: Pos = field(
        default_factory=lambda: Pos(), metadata={"description": "Text position"}
    )
    font: Optional[Font] = field(
        default=None, metadata={"description": "Font settings", "required": False}
    )
    justify: Optional[Justify] = field(
        default=None, metadata={"description": "Text justification", "required": False}
    )
    repeat: KiCadInt = field(
        default_factory=lambda: KiCadInt("repeat", 0, required=False),
        metadata={"description": "Repeat count", "required": False},
    )
    incrx: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incrx", 0.0, required=False),
        metadata={"description": "X increment", "required": False},
    )
    incry: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incry", 0.0, required=False),
        metadata={"description": "Y increment", "required": False},
    )
    comment: KiCadStr = field(
        default_factory=lambda: KiCadStr("comment", "", required=False),
        metadata={"description": "Comment", "required": False},
    )


@dataclass
class KicadWks(KiCadObject):
    """KiCad worksheet definition token.

    The 'kicad_wks' token defines worksheet format information in the format::

        (kicad_wks
            (version VERSION)
            (generator GENERATOR)
            ;; contents of the schematic file...
        )

    Args:
        version: Format version (optional)
        generator: Generator name (optional)
        generator_version: Generator version (optional)
        page: Page settings (optional)
        title_block: Title block (optional)
        setup: Worksheet setup (optional)
        rect: List of rectangles (optional)
        line: List of lines (optional)
        tbtext: List of text blocks (optional)
        elements: List of worksheet elements (optional)
    """

    __token_name__ = "kicad_wks"
    __legacy_token_names__ = ["page_layout"]

    version: KiCadInt = field(
        default_factory=lambda: KiCadInt("version", 0, required=False),
        metadata={"description": "Format version", "required": False},
    )
    generator: KiCadStr = field(
        default_factory=lambda: KiCadStr("generator", "", required=False),
        metadata={"description": "Generator name", "required": False},
    )
    generator_version: KiCadStr = field(
        default_factory=lambda: KiCadStr("generator_version", "", required=False),
        metadata={"description": "Generator version", "required": False},
    )
    page: KiCadStr = field(
        default_factory=lambda: KiCadStr("page", "", required=False),
        metadata={"description": "Page settings", "required": False},
    )
    title_block: Optional[TitleBlock] = field(
        default=None, metadata={"description": "Title block", "required": False}
    )
    setup: Optional[WksSetup] = field(
        default=None, metadata={"description": "Worksheet setup", "required": False}
    )
    rect: Optional[List[WksRect]] = field(
        default_factory=list,
        metadata={"description": "List of rectangles", "required": False},
    )
    line: Optional[List[WksLine]] = field(
        default_factory=list,
        metadata={"description": "List of lines", "required": False},
    )
    tbtext: Optional[List[WksTbText]] = field(
        default_factory=list,
        metadata={"description": "List of text blocks", "required": False},
    )
    elements: Optional[List[Any]] = field(
        default_factory=list,
        metadata={"description": "List of worksheet elements", "required": False},
    )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        strictness: ParseStrictness = ParseStrictness.STRICT,
        encoding: str = "utf-8",
    ) -> "KicadWks":
        """Parse from S-expression file - convenience method for worksheet operations."""
        if not file_path.endswith(".kicad_wks"):
            raise ValueError("Unsupported file extension. Expected: .kicad_wks")
        with open(file_path, "r", encoding=encoding) as f:
            content = f.read()
        return cls.from_str(content, strictness)

    def save_to_file(self, file_path: str, encoding: str = "utf-8") -> None:
        """Save to .kicad_wks file format.

        Args:
            file_path: Path to write the .kicad_wks file
            encoding: File encoding (default: utf-8)
        """
        if not file_path.endswith(".kicad_wks"):
            raise ValueError("Unsupported file extension. Expected: .kicad_wks")
        content = self.to_sexpr_str()
        with open(file_path, "w", encoding=encoding) as f:
            f.write(content)


# Image related elements
@dataclass
class Bitmap(KiCadObject):
    """Bitmap image definition token.

    The 'bitmap' token defines a bitmap image in the format::

        (bitmap
            (name "NAME")
            (pos X Y)
            (scale SCALAR)
            [(repeat COUNT)]
            [(incrx DISTANCE)]
            [(incry DISTANCE)]
            [(comment "COMMENT")]
            (pngdata IMAGE_DATA)
        )

    Args:
        name: Image name
        pos: Position coordinates
        scale: Scale factor
        repeat: Repeat count (optional)
        incrx: X increment distance (optional)
        incry: Y increment distance (optional)
        comment: Image comment (optional)
        pngdata: PNG image data
    """

    __token_name__ = "bitmap"

    name: KiCadStr = field(
        default_factory=lambda: KiCadStr("name", ""),
        metadata={"description": "Image name"},
    )
    pos: Pos = field(
        default_factory=lambda: Pos(), metadata={"description": "Position coordinates"}
    )
    scale: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("factor", 0.0),
        metadata={"description": "Scale factor"},
    )
    repeat: KiCadInt = field(
        default_factory=lambda: KiCadInt("repeat", 0, required=False),
        metadata={"description": "Repeat count", "required": False},
    )
    incrx: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incrx", 0.0, required=False),
        metadata={"description": "X increment distance", "required": False},
    )
    incry: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("incry", 0.0, required=False),
        metadata={"description": "Y increment distance", "required": False},
    )
    comment: Optional[str] = field(
        default=None, metadata={"description": "Image comment", "required": False}
    )
    pngdata: "Pngdata" = field(
        default_factory=lambda: Pngdata(), metadata={"description": "PNG image data"}
    )


@dataclass
class Image(KiCadObject):
    """Image definition token.

    The 'image' token defines an image object in PCB files in the format::

        (image (at X Y) (scale FACTOR) (uuid UUID) (data ...))

    Args:
        at: Position
        scale: Scale factor (optional)
        uuid: Unique identifier (optional)
        data: Image data (optional)
        locked: Whether image is locked (optional)
    """

    __token_name__ = "image"

    at: AtXY = field(
        default_factory=lambda: AtXY(), metadata={"description": "Position"}
    )
    scale: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("scale", 1.0, required=False),
        metadata={"description": "Scale factor", "required": False},
    )
    uuid: Optional[Uuid] = field(
        default=None, metadata={"description": "Unique identifier", "required": False}
    )
    data: Optional[Data] = field(
        default=None, metadata={"description": "Image data", "required": False}
    )
    locked: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("locked"),
        metadata={"description": "Whether image is locked", "required": False},
    )


@dataclass
class Pngdata(KiCadObject):
    """PNG data definition token.

    The 'pngdata' token defines PNG image data in the format::

        (pngdata
            (data XX1 ... XXN)
            (data XX1 ... XXN)
            ...
        )

    Where each data line contains up to 32 hexadecimal bytes.

    Args:
        data_lines: List of data token objects containing hexadecimal bytes
    """

    __token_name__ = "pngdata"

    data_lines: List[Data] = field(
        default_factory=list,
        metadata={
            "description": "List of data token objects containing hexadecimal bytes"
        },
    )



"""Base types for KiCad S-expressions - fundamental elements with no cross-dependencies."""

from dataclasses import dataclass, field
from typing import Any, List, Optional

from .base_element import KiCadFloat, KiCadObject, KiCadStr, OptionalFlag
from .enums import PadShape, StrokeType
from .sexpr_parser import SExpr


@dataclass
class Anchor(KiCadObject):
    """Anchor pad shape definition for custom pads.

    The 'anchor' token defines the anchor pad shape of a custom pad in the format::

        (anchor PAD_SHAPE)

    Args:
        pad_shape: Anchor pad shape (rect or circle)
    """

    __token_name__ = "anchor"

    pad_shape: PadShape = field(
        default=PadShape.RECT,
        metadata={"description": "Anchor pad shape (rect or circle)"},
    )


@dataclass
class Xy(KiCadObject):
    """2D coordinate definition token.

    The 'xy' token defines a 2D coordinate point in the format:
    (xy X Y)

    Args:
        x: Horizontal coordinate
        y: Vertical coordinate
    """

    __token_name__ = "xy"

    x: float = field(default=0.0, metadata={"description": "Horizontal coordinate"})
    y: float = field(default=0.0, metadata={"description": "Vertical coordinate"})


@dataclass
class Xyz(KiCadObject):
    """3D coordinate definition token.

    The 'xyz' token defines 3D coordinates in the format:
    (xyz X Y Z)

    Args:
        x: X coordinate
        y: Y coordinate
        z: Z coordinate
    """

    __token_name__ = "xyz"

    x: float = field(default=0.0, metadata={"description": "X coordinate"})
    y: float = field(default=0.0, metadata={"description": "Y coordinate"})
    z: float = field(default=0.0, metadata={"description": "Z coordinate"})


@dataclass
class Pts(KiCadObject):
    """Coordinate point list definition token.

    The 'pts' token defines a list of coordinate points in the format:
    (pts
        (xy X Y)
        ...
        (xy X Y)
    )

    Where each xy token defines a single X and Y coordinate pair.
    The number of points is determined by the object type.

    Args:
        points: List of 2D coordinate points
    """

    __token_name__ = "pts"

    points: List[Xy] = field(
        default_factory=list, metadata={"description": "List of 2D coordinate points"}
    )

    def xy(self, x: float, y: float) -> "Pts":
        """Add a coordinate point to the list.

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            Pts: Self (for method chaining)
        """
        self.points.append(Xy(x=x, y=y))
        return self


@dataclass
class AtXY(KiCadObject):
    """Position identifier token for elements that only use X and Y coordinates.

    The 'at' token defines positional coordinates in the format:
        (at X Y)

    Used for elements like junctions that don't have rotation.

    Args:
        x: Horizontal position of the object
        y: Vertical position of the object
    """

    __token_name__ = "at"

    x: float = field(
        default=0.0,
        metadata={"description": "Horizontal position of the object"},
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the object"}
    )

    def xy(self, x: float, y: float) -> "AtXY":
        """Set X and Y coordinates.

        Args:
            x: Horizontal position
            y: Vertical position

        Returns:
            AtXY: Self (for method chaining)
        """
        self.x = x
        self.y = y
        return self


@dataclass
class At(KiCadObject):
    """Position identifier token that defines positional coordinates and rotation of an object.

    The 'at' token defines the positional coordinates in the format:
        (at X Y [ANGLE])

    Note:
        Symbol text ANGLEs are stored in tenth's of a degree. All other ANGLEs are stored in degrees.
        For 3D model positioning, use ModelAt class which supports (at (xyz X Y Z)) format.

    Args:
        x: Horizontal position of the object
        y: Vertical position of the object
        angle: Rotational angle of the object
    """

    __token_name__ = "at"

    x: float = field(
        default=0.0,
        metadata={"description": "Horizontal position of the object"},
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the object"}
    )
    angle: Optional[float] = field(
        default=0.0,
        metadata={"description": "Rotational angle of the object"},
    )


@dataclass
class Center(KiCadObject):
    """Center point definition token.

    The 'center' token defines a center point in the format::

        (center X Y)

    Args:
        x: Horizontal position of the center point
        y: Vertical position of the center point
    """

    __token_name__ = "center"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal position of the center point"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the center point"}
    )


@dataclass
class Color(KiCadObject):
    """Color definition token.

    The 'color' token defines color values in the format::

        (color R G B A)

    Args:
        r: Red color component (0-255)
        g: Green color component (0-255)
        b: Blue color component (0-255)
        a: Alpha component (0-255)
    """

    __token_name__ = "color"

    r: int = field(default=0, metadata={"description": "Red color component (0-255)"})
    g: int = field(default=0, metadata={"description": "Green color component (0-255)"})
    b: int = field(default=0, metadata={"description": "Blue color component (0-255)"})
    a: int = field(default=0, metadata={"description": "Alpha component (0-255)"})


@dataclass
class End(KiCadObject):
    """End point definition token.

    The 'end' token defines an end point in the format::

        (end X Y)

    Args:
        x: Horizontal position of the end point
        y: Vertical position of the end point
        corner: Corner reference (optional)
    """

    __token_name__ = "end"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal position of the end point"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the end point"}
    )
    corner: Optional[str] = field(
        default=None, metadata={"description": "Corner reference", "required": False}
    )


@dataclass
class Mid(KiCadObject):
    """Mid point definition token.

    The 'mid' token defines a mid point in the format::

        (mid X Y)

    Args:
        x: Horizontal position of the mid point
        y: Vertical position of the mid point
    """

    __token_name__ = "mid"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal position of the mid point"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the mid point"}
    )


@dataclass
class Type(KiCadObject):
    """Type definition token.

    The 'type' token defines a type value in the format:
    (type VALUE)

    Args:
        value: Type value
    """

    __token_name__ = "type"

    value: str = field(default="", metadata={"description": "Type value"})

    def to_sexpr(self) -> SExpr:
        """Custom serialization to ensure type values are never quoted."""
        return [self.__token_name__, self.value]


@dataclass
class Fill(KiCadObject):
    """Fill definition token.

    The 'fill' token defines how schematic and symbol library graphical items are filled in the format:
    (fill
        (type none | outline | background)
        (color R G B A)
    )

    This represents the nested structure exactly as it appears in the S-expression files.

    Args:
        type: Fill type specification (optional)
        color: Fill color specification (optional)
    """

    __token_name__ = "fill"

    type: Optional[Type] = field(
        default=None,
        metadata={"description": "Fill type specification", "required": False},
    )
    color: Optional[Color] = field(
        default=None,
        metadata={"description": "Fill color specification", "required": False},
    )


@dataclass
class Layer(KiCadObject):
    """Layer definition token.

    The 'layer' token defines layer information in the format::

        (layer "NAME" | dielectric NUMBER (type "DESCRIPTION")
               [(color "COLOR")] [(thickness THICKNESS)]
               [(material "MATERIAL")] [(epsilon_r VALUE)]
               [(loss_tangent VALUE)])

    For simple layer references:
        (layer "LAYER_NAME")

    Args:
        name: Layer name or 'dielectric'
        number: Layer stack number (optional)
        type: Layer type description (optional)
        color: Layer color as string (optional)
        thickness: Layer thickness value (optional)
        material: Material name (optional)
        epsilon_r: Dielectric constant value (optional)
        loss_tangent: Loss tangent value (optional)
    """

    __token_name__ = "layer"

    name: str = field(
        default="", metadata={"description": "Layer name or 'dielectric'"}
    )
    number: Optional[int] = field(
        default=None, metadata={"description": "Layer stack number", "required": False}
    )
    type: Optional[str] = field(
        default=None,
        metadata={"description": "Layer type description", "required": False},
    )
    color: Optional[str] = field(
        default=None,
        metadata={"description": "Layer color as string", "required": False},
    )
    thickness: Optional[float] = field(
        default=None,
        metadata={"description": "Layer thickness value", "required": False},
    )
    material: Optional[str] = field(
        default=None, metadata={"description": "Material name", "required": False}
    )
    epsilon_r: Optional[float] = field(
        default=None,  # 4,5 nomally
        metadata={"description": "Dielectric constant value", "required": False},
    )
    loss_tangent: Optional[float] = field(
        default=None, metadata={"description": "Loss tangent value", "required": False}
    )


@dataclass
class Offset(KiCadObject):
    """Offset definition token.

    The 'offset' token defines an offset position in the format:
    (offset X Y)

    Args:
        x: Horizontal offset coordinate
        y: Vertical offset coordinate
    """

    __token_name__ = "offset"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal offset coordinate"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical offset coordinate"}
    )


@dataclass
class Pos(KiCadObject):
    """Position definition token.

    The 'pos' token defines a position in the format:
    (pos X Y)

    Args:
        x: Horizontal position coordinate
        y: Vertical position coordinate
        corner: Corner reference (optional)
    """

    __token_name__ = "pos"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal position coordinate"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position coordinate"}
    )
    corner: Optional[str] = field(
        default=None, metadata={"description": "Corner reference", "required": False}
    )


@dataclass
class Size(KiCadObject):
    """Size definition token.

    The 'size' token defines width and height dimensions in the format:
    (size WIDTH HEIGHT)

    Args:
        width: Width dimension
        height: Height dimension
    """

    __token_name__ = "size"

    width: float = field(default=0.0, metadata={"description": "Width dimension"})
    height: float = field(default=0.0, metadata={"description": "Height dimension"})


@dataclass
class Start(KiCadObject):
    """Start point definition token.

    The 'start' token defines a start point in the format:
    (start X Y)

    Args:
        x: Horizontal position of the start point
        y: Vertical position of the start point
        corner: Corner reference (optional)
    """

    __token_name__ = "start"

    x: float = field(
        default=0.0, metadata={"description": "Horizontal position of the start point"}
    )
    y: float = field(
        default=0.0, metadata={"description": "Vertical position of the start point"}
    )
    corner: Optional[str] = field(
        default=None, metadata={"description": "Corner reference", "required": False}
    )


@dataclass
class Stroke(KiCadObject):
    """Stroke definition token.

    The 'stroke' token defines how the outlines of graphical objects are drawn in the format:
    (stroke
        (width WIDTH)
        (type TYPE)
        (color R G B A)
    )

    This represents the nested structure exactly as it appears in the S-expression files.

    Args:
        width: Line width specification
        type: Stroke line style specification
        color: Line color specification (optional)
    """

    __token_name__ = "stroke"

    width: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("width", 0.0),
        metadata={"description": "Line width specification"},
    )
    type: Type = field(
        default_factory=lambda: Type(value=StrokeType.SOLID.value),
        metadata={"description": "Stroke line style specification"},
    )
    color: Optional[Color] = field(
        default=None,
        metadata={"description": "Line color specification", "required": False},
    )


@dataclass
class Uuid(KiCadObject):
    """UUID identifier token.

    The 'uuid' token defines a universally unique identifier in the format:
    (uuid UUID_VALUE)

    Args:
        value: UUID value
    """

    __token_name__ = "uuid"

    value: str = field(default="", metadata={"description": "UUID value"})

    def new_id(self) -> "Uuid":
        """Generate a new UUID identifier and update this instance.

        Returns:
            Uuid: Self (for method chaining)
        """
        import uuid

        self.value = str(uuid.uuid4())
        return self

    @classmethod
    def create(cls) -> "Uuid":
        """Create a new UUID identifier (factory method).

        Returns:
            Uuid: New Uuid object with a generated UUID value
        """
        import uuid

        return cls(value=str(uuid.uuid4()))


@dataclass
class Font(KiCadObject):
    """Font definition token.

    The 'font' token defines font properties in the format:
    (font [face "FONT_NAME"] [size WIDTH HEIGHT] [thickness THICKNESS] [bold] [italic])

    Args:
        face: Font face specification (optional)
        size: Font size (optional)
        thickness: Font thickness (optional)
        bold: Bold flag (optional)
        italic: Italic flag (optional)
        color: Font color (optional)
    """

    __token_name__ = "font"

    face: KiCadStr = field(
        default_factory=lambda: KiCadStr("face", "", required=False),
        metadata={"description": "Font face specification", "required": False},
    )
    size: Optional[Size] = field(
        default_factory=lambda: Size(1.27, 1.27),
        metadata={"description": "Font size", "required": False},
    )
    thickness: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("thickness", 0.0, required=False),
        metadata={"description": "Font thickness", "required": False},
    )
    bold: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("bold"),
        metadata={"description": "Bold flag", "required": False},
    )
    italic: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("italic"),
        metadata={"description": "Italic flag", "required": False},
    )
    color: Optional[Color] = field(
        default=None,
        metadata={"description": "Font color", "required": False},
    )


@dataclass
class Justify(KiCadObject):
    """Text justification definition token.

    The 'justify' token defines text alignment and mirroring in the format::

        (justify [left | right | center] [top | bottom | center] [mirror])

    Args:
        left: Left horizontal justification flag (optional)
        right: Right horizontal justification flag (optional)
        top: Top vertical justification flag (optional)
        bottom: Bottom vertical justification flag (optional)
        center: Center justification flag (horizontal or vertical) (optional)
        mirror: Mirror text flag (optional)
    """

    __token_name__ = "justify"

    # Horizontal justification flags
    left: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("left"),
        metadata={
            "description": "Left horizontal justification flag",
            "required": False,
        },
    )
    right: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("right"),
        metadata={
            "description": "Right horizontal justification flag",
            "required": False,
        },
    )

    # Vertical justification flags
    top: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("top"),
        metadata={"description": "Top vertical justification flag", "required": False},
    )
    bottom: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("bottom"),
        metadata={
            "description": "Bottom vertical justification flag",
            "required": False,
        },
    )

    # Center can be horizontal or vertical - ambiguous in S-expression
    center: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("center"),
        metadata={
            "description": "Center justification flag (horizontal or vertical)",
            "required": False,
        },
    )

    # Mirror flag
    mirror: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("mirror"),
        metadata={"description": "Mirror text flag", "required": False},
    )


@dataclass
class Effects(KiCadObject):
    """Text effects definition token.

    The 'effects' token defines text formatting effects in the format::

        (effects
            (font [size SIZE] [thickness THICKNESS] [bold] [italic])
            [(justify JUSTIFY)]
            [hide]
        )

    Args:
        font: Font definition (optional)
        justify: Text justification (optional)
        hide: Whether text is hidden (optional)
        href: Hyperlink reference (optional)
    """

    __token_name__ = "effects"

    font: Optional[Font] = field(
        default=None, metadata={"description": "Font definition", "required": False}
    )
    justify: Optional[Justify] = field(
        default=None,
        metadata={"description": "Text justification", "required": False},
    )
    hide: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("hide"),
        metadata={"description": "Whether text is hidden", "required": False},
    )
    href: KiCadStr = field(
        default_factory=lambda: KiCadStr("href", "", required=False),
        metadata={"description": "Hyperlink reference", "required": False},
    )


@dataclass
class Text(KiCadObject):
    """Text content definition token.

    The 'text' token defines text content in the format:
    (text "TEXT_CONTENT"
        (at X Y [ANGLE])
        (effects EFFECTS)
    )

    Args:
        content: Text content
        at: Position and rotation (optional)
        effects: Text effects (optional)
        exclude_from_sim: Whether to exclude from simulation (optional)
        uuid: Unique identifier (optional)
    """

    __token_name__ = "text"

    content: str = field(default="", metadata={"description": "Text content"})
    at: Optional[At] = field(
        default=None,
        metadata={"description": "Position and rotation", "required": False},
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    exclude_from_sim: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("exclude_from_sim"),
        metadata={
            "description": "Whether to exclude from simulation",
            "required": False,
        },
    )
    uuid: Optional[Uuid] = field(
        default=None,
        metadata={"description": "Unique identifier", "required": False},
    )


@dataclass
class Property(KiCadObject):
    """Property definition token.

    The 'property' token defines properties in two formats:

    General properties::
        (property "KEY" "VALUE")

    Symbol properties::
        (property
            "KEY"
            "VALUE"
            (id N)
            POSITION_IDENTIFIER
            TEXT_EFFECTS
        )

    Args:
        key: Property key name (must be unique)
        value: Property value
        id: Property ID (optional)
        at: Position and rotation (optional)
        effects: Text effects (optional)
        unlocked: Whether property is unlocked (optional)
        layer: Layer assignment (optional)
        uuid: Unique identifier (optional)
        hide: Hide property flag (optional)
    """

    __token_name__ = "property"

    key: str = field(
        default="", metadata={"description": "Property key name (must be unique)"}
    )
    value: str = field(default="", metadata={"description": "Property value"})
    id: KiCadStr = field(
        default_factory=lambda: KiCadStr("id", "", required=False),
        metadata={"description": "Property ID", "required": False},
    )
    at: Optional[At] = field(
        default=None,
        metadata={"description": "Position and rotation", "required": False},
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    unlocked: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("unlocked"),
        metadata={"description": "Whether property is unlocked", "required": False},
    )
    layer: Optional[Layer] = field(
        default=None, metadata={"description": "Layer assignment", "required": False}
    )
    uuid: Optional[Uuid] = field(
        default=None, metadata={"description": "Unique identifier", "required": False}
    )
    hide: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("hide"),
        metadata={"description": "Hide property flag", "required": False},
    )


@dataclass
class LayerDefinition:
    """Board layer definition entry (no token name).

    Individual layer definition within a 'layers' token in PCB files::

        (0 "F.Cu" signal)
        (9 "F.Adhes" user "F.Adhesive")
        (31 "F.CrtYd" user "F.Courtyard")

    Stores the raw list internally and provides properties for access.

    Args:
        data: Raw layer data as list [ordinal, name, type, user_name?]
    """

    data: List[Any] = field(default_factory=list)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize from list or individual parameters."""
        if len(args) == 1 and isinstance(args[0], list):
            self.data = args[0]
        elif args:
            self.data = list(args)
        else:
            self.data = []

    @property
    def ordinal(self) -> int:
        """Layer ordinal number."""
        return int(self.data[0]) if len(self.data) > 0 else 0

    @property
    def canonical_name(self) -> str:
        """Canonical layer name."""
        return self.data[1] if len(self.data) > 1 else ""

    @property
    def layer_type(self) -> str:
        """Layer type (signal, user, power, mixed, jumper)."""
        return self.data[2] if len(self.data) > 2 else "signal"

    @property
    def user_name(self) -> Optional[str]:
        """User-defined layer name (optional)."""
        return self.data[3] if len(self.data) > 3 else None


@dataclass
class BoardLayers(KiCadObject):
    """Board layer definitions for kicad_pcb files.

    Stores layer definitions as raw lists::

        (layers (0 "F.Cu" signal) (2 "B.Cu" signal) ...)

    Use get_layer(index) to parse a specific layer as LayerDefinition if needed.
    """

    __token_name__ = "layers"
    layer_defs: List[LayerDefinition] = field(default_factory=list)

    @classmethod
    def _parse_recursive(cls, cursor: Any) -> "BoardLayers":
        """Parse and store layer definitions as LayerDefinition objects."""
        layer_defs = [
            LayerDefinition([str(v) for v in item])
            for item in cursor.sexpr[1:]
            if isinstance(item, list)
        ]
        for i in range(1, len(cursor.sexpr)):
            cursor.parser.mark_used(i)
        return cls(layer_defs=layer_defs)

    def to_sexpr(self) -> SExpr:
        """Convert to S-expression format."""
        return [self.__token_name__] + [ld.data for ld in self.layer_defs]


@dataclass
class Layers(KiCadObject):
    """Layer list definition token.

    The 'layers' token defines a list of layer names in the format::

        (layers "F.Cu" "F.Paste" "F.Mask")
        (layers "*.Cu" "*.Mask" "F.SilkS")

    Used for pad layers, via layers, and other layer specifications.

    Attributes:
        layers (List[str]): List of layer names

    Args:
        layers: List of layer names
    """

    __token_name__ = "layers"

    layers: List[str] = field(
        default_factory=list, metadata={"description": "List of layer names"}
    )



"""Schematic system elements for KiCad S-expressions - schematic drawing and connectivity."""

from dataclasses import dataclass, field
from typing import Any, List, Optional, Union

from .base_element import (
    KiCadFloat,
    KiCadInt,
    KiCadObject,
    KiCadStr,
    OptionalFlag,
    ParseStrictness,
)
from .base_types import (
    At,
    AtXY,
    Color,
    Effects,
    Fill,
    Font,
    Justify,
    Property,
    Pts,
    Size,
    Stroke,
    Text,
    Uuid,
)
from .enums import LabelShape
from .primitive_graphics import Arc, Bezier, Circle, Line, Polygon, Polyline, Rectangle
from .symbol_library import LibSymbols
from .text_and_documents import (
    Image,
    Paper,
    TitleBlock,
)


@dataclass
class TableBorder(KiCadObject):
    """Table border definition token.

    The 'border' token defines border configuration for tables in the format::
        (border (external yes) (header yes) (stroke (width WIDTH) (type TYPE)))

    Args:
        external: Whether external border is shown (optional)
        header: Whether header border is shown (optional)
        stroke: Stroke definition for border lines (optional)
    """

    __token_name__ = "border"

    external: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("external"),
        metadata={"description": "Whether external border is shown", "required": False},
    )
    header: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("header"),
        metadata={"description": "Whether header border is shown", "required": False},
    )
    stroke: Optional[Stroke] = field(
        default=None,
        metadata={
            "description": "Stroke definition for border lines",
            "required": False,
        },
    )


@dataclass
class TableSeparators(KiCadObject):
    """Table separators definition token.

    The 'separators' token defines separator configuration for tables in the format::
        (separators (rows yes) (cols yes) (stroke (width WIDTH) (type TYPE)))

    Args:
        rows: Whether row separators are shown (optional)
        cols: Whether column separators are shown (optional)
        stroke: Stroke definition for separator lines (optional)
    """

    __token_name__ = "separators"

    rows: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("rows"),
        metadata={"description": "Whether row separators are shown", "required": False},
    )
    cols: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("cols"),
        metadata={
            "description": "Whether column separators are shown",
            "required": False,
        },
    )
    stroke: Optional[Stroke] = field(
        default=None,
        metadata={
            "description": "Stroke definition for separator lines",
            "required": False,
        },
    )


@dataclass
class TableMargins(KiCadObject):
    """Table margins definition token.

    The 'margins' token defines cell margins in the format::
        (margins LEFT TOP RIGHT BOTTOM)

    Args:
        left: Left margin
        top: Top margin
        right: Right margin
        bottom: Bottom margin
    """

    __token_name__ = "margins"

    left: float = field(default=0.0, metadata={"description": "Left margin"})
    top: float = field(default=0.0, metadata={"description": "Top margin"})
    right: float = field(default=0.0, metadata={"description": "Right margin"})
    bottom: float = field(default=0.0, metadata={"description": "Bottom margin"})


@dataclass
class TableSpan(KiCadObject):
    """Table span definition token.

    The 'span' token defines cell span in the format::
        (span COLS ROWS)

    Args:
        cols: Number of columns to span
        rows: Number of rows to span
    """

    __token_name__ = "span"

    cols: int = field(default=1, metadata={"description": "Number of columns to span"})
    rows: int = field(default=1, metadata={"description": "Number of rows to span"})


@dataclass
class TableCell(KiCadObject):
    """Table cell definition token.

    The 'table_cell' token defines individual table cells in the format::
        (table_cell "TEXT"
            (exclude_from_sim BOOLEAN)
            (at X Y ANGLE)
            (size WIDTH HEIGHT)
            (margins LEFT TOP RIGHT BOTTOM)
            (span COLS ROWS)
            (fill FILL_DEF)
            (effects EFFECTS_DEF)
            (uuid UUID)
        )

    Args:
        text: Cell text content
        exclude_from_sim: Whether to exclude from simulation (optional)
        at: Position and rotation (optional)
        size: Cell size (optional)
        margins: Cell margins (optional)
        span: Cell span (optional)
        fill: Fill definition (optional)
        effects: Text effects (optional)
        uuid: Unique identifier (optional)
    """

    __token_name__ = "table_cell"

    text: str = field(default="", metadata={"description": "Cell text content"})
    exclude_from_sim: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("exclude_from_sim"),
        metadata={
            "description": "Whether to exclude from simulation",
            "required": False,
        },
    )
    at: Optional[At] = field(
        default=None,
        metadata={"description": "Position and rotation", "required": False},
    )
    size: Optional[Size] = field(
        default=None,
        metadata={"description": "Cell size", "required": False},
    )
    margins: Optional[TableMargins] = field(
        default=None,
        metadata={"description": "Cell margins", "required": False},
    )
    span: Optional[TableSpan] = field(
        default=None,
        metadata={"description": "Cell span", "required": False},
    )
    fill: Optional[Fill] = field(
        default=None,
        metadata={"description": "Fill definition", "required": False},
    )
    effects: Optional[Effects] = field(
        default=None,
        metadata={"description": "Text effects", "required": False},
    )
    uuid: Optional[Uuid] = field(
        default=None,
        metadata={"description": "Unique identifier", "required": False},
    )


@dataclass
class RowHeights(KiCadObject):
    """Row heights definition token.

    The 'row_heights' token defines row height values in the format::
        (row_heights HEIGHT1 HEIGHT2 ...)

    Args:
        heights: List of row height values
    """

    __token_name__ = "row_heights"

    heights: List[float] = field(
        default_factory=list,
        metadata={"description": "List of row height values"},
    )


@dataclass
class ColumnWidths(KiCadObject):
    """Column widths definition token.

    The 'column_widths' token defines column width values in the format::
        (column_widths WIDTH1 WIDTH2 ...)

    Args:
        widths: List of column width values
    """

    __token_name__ = "column_widths"

    widths: List[float] = field(
        default_factory=list,
        metadata={"description": "List of column width values"},
    )


@dataclass
class Cells(KiCadObject):
    """Table cells container token.

    The 'cells' token contains a list of table_cell objects in the format::
        (cells
            (table_cell ...)
            (table_cell ...)
            ...
        )

    Args:
        cells: List of table cell objects
    """

    __token_name__ = "cells"

    cells: List[TableCell] = field(
        default_factory=list,
        metadata={"description": "List of table cell objects"},
    )


@dataclass
class Table(KiCadObject):
    """Table definition token for schematics.

    The 'table' token defines tables in schematics.
    This is a basic implementation for parsing support.

    Args:
        column_count: Number of columns (optional)
        border: Border configuration (optional)
        separators: Separator configuration (optional)
        column_widths: Column width values (optional)
        row_heights: Row height values (optional)
        cells: Table cells (optional)
    """

    __token_name__ = "table"

    column_count: Optional[int] = field(
        default=None,
        metadata={"description": "Number of columns", "required": False},
    )
    border: Optional[TableBorder] = field(
        default=None,
        metadata={"description": "Border configuration", "required": False},
    )
    separators: Optional[TableSeparators] = field(
        default=None,
        metadata={"description": "Separator configuration", "required": False},
    )
    column_widths: Optional[ColumnWidths] = field(
        default=None,
        metadata={"description": "Column width values", "required": False},
    )
    row_heights: Optional[RowHeights] = field(
        default=None,
        metadata={"description": "Row height values", "required": False},
    )
    cells: Optional[Cells] = field(
        default=None,
        metadata={"description": "Table cells", "required": False},
    )


@dataclass
class HierarchicalLabel(KiCadObject):
    """Hierarchical label definition token.

    The 'hierarchical_label' token defines hierarchical labels in schematics.

    Args:
        text: Label text
        shape: Label shape (optional)
        at: Position (optional)
        effects: Text effects (optional)
        uuid: Unique identifier (optional)
    """

    __token_name__ = "hierarchical_label"

    text: str = field(default="", metadata={"description": "Label text"})
    shape: Optional[LabelShape] = field(
        default=None, metadata={"description": "Label shape", "required": False}
    )
    at: Optional[At] = field(
        default=None, metadata={"description": "Position", "required": False}
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    uuid: Optional[Uuid] = field(
        default=None, metadata={"description": "Unique identifier", "required": False}
    )


@dataclass
class RuleArea(KiCadObject):
    """Rule area definition token.

    The 'rule_area' token defines rule areas in schematics.

    Args:
        polylines: Polylines defining area (optional)
    """

    __token_name__ = "rule_area"

    polylines: Optional[List[Polyline]] = field(
        default_factory=list,
        metadata={"description": "Polylines defining area", "required": False},
    )


@dataclass
class NetclassFlag(KiCadObject):
    """Netclass flag definition token.

    The 'netclass_flag' token defines netclass flags in schematics.

    Args:
        name: Netclass name
        length: Flag length (optional)
        shape: Flag shape (optional)
        at: Position (optional)
        fields_autoplaced: Whether fields are auto-placed (optional)
        effects: Text effects (optional)
        uuid: Unique identifier (optional)
        properties: Properties of the netclass flag (optional)
    """

    __token_name__ = "netclass_flag"

    name: str = field(default="", metadata={"description": "Netclass name"})
    length: Optional[float] = field(
        default=None, metadata={"description": "Flag length", "required": False}
    )
    shape: Optional[str] = field(
        default=None, metadata={"description": "Flag shape", "required": False}
    )
    at: Optional[At] = field(
        default=None, metadata={"description": "Position", "required": False}
    )
    fields_autoplaced: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("fields_autoplaced"),
        metadata={"description": "Whether fields are auto-placed", "required": False},
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    uuid: Optional[Uuid] = field(
        default=None, metadata={"description": "Unique identifier", "required": False}
    )
    properties: Optional[List[Property]] = field(
        default_factory=list,
        metadata={"description": "Properties of the netclass flag", "required": False},
    )


@dataclass
class Bus(KiCadObject):
    """Bus definition token.

    The 'bus' token defines buses in the schematic in the format::

        (bus
            COORDINATE_POINT_LIST
            STROKE_DEFINITION
            UNIQUE_IDENTIFIER
        )

    Args:
        pts: Bus connection points
        stroke: Stroke definition
        uuid: Unique identifier
    """

    __token_name__ = "bus"

    pts: Pts = field(
        default_factory=lambda: Pts(), metadata={"description": "Bus connection points"}
    )
    stroke: Stroke = field(
        default_factory=lambda: Stroke(), metadata={"description": "Stroke definition"}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class BusEntry(KiCadObject):
    """Bus entry definition token.

    The 'bus_entry' token defines a bus entry in the schematic in the format::

        (bus_entry
            POSITION_IDENTIFIER
            (size X Y)
            STROKE_DEFINITION
            UNIQUE_IDENTIFIER
        )

    Args:
        at: Position
        size: Entry size
        stroke: Stroke definition
        uuid: Unique identifier
    """

    __token_name__ = "bus_entry"

    at: AtXY = field(
        default_factory=lambda: AtXY(), metadata={"description": "Position"}
    )
    size: Size = field(
        default_factory=lambda: Size(), metadata={"description": "Entry size"}
    )
    stroke: Stroke = field(
        default_factory=lambda: Stroke(), metadata={"description": "Stroke definition"}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class GlobalLabel(KiCadObject):
    """Global label definition token.

    The 'global_label' token defines a label visible across all schematics in the format::

        (global_label
            "TEXT"
            (shape SHAPE)
            [(fields_autoplaced)]
            POSITION_IDENTIFIER
            TEXT_EFFECTS
            UNIQUE_IDENTIFIER
            PROPERTIES
        )

    Args:
        text: Global label text
        shape: Way the global label is drawn (optional)
        fields_autoplaced: Whether properties are placed automatically (optional)
        at: X and Y coordinates and rotation angle
        effects: How the global label text is drawn (optional)
        uuid: Universally unique identifier
        properties: Properties of the global label (optional)
    """

    __token_name__ = "global_label"

    text: str = field(default="", metadata={"description": "Global label text"})
    shape: KiCadStr = field(
        default_factory=lambda: KiCadStr("shape", "bidirectional"),
        metadata={"description": "Way the global label is drawn", "required": False},
    )
    fields_autoplaced: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("fields_autoplaced"),
        metadata={
            "description": "Whether properties are placed automatically",
            "required": False,
        },
    )
    at: At = field(
        default_factory=lambda: At(),
        metadata={"description": "X and Y coordinates and rotation angle"},
    )
    effects: Optional[Effects] = field(
        default_factory=lambda: Effects(font=Font(), justify=Justify()),
        metadata={
            "description": "How the global label text is drawn",
            "required": False,
        },
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(),
        metadata={"description": "Universally unique identifier"},
    )
    properties: Optional[List[Property]] = field(
        default_factory=lambda: [
            Property(
                key="Intersheetrefs",
                value="${INTERSHEET_REFS}",
                effects=Effects(font=Font(), hide=OptionalFlag("hide")),
            )
        ],
        metadata={"description": "Properties of the global label", "required": False},
    )


@dataclass
class Junction(KiCadObject):
    """Junction definition token.

    The 'junction' token defines a junction in the schematic in the format::

        (junction
            POSITION_IDENTIFIER
            (diameter DIAMETER)
            (color R G B A)
            UNIQUE_IDENTIFIER
        )

    Args:
        at: Position
        diameter: Junction diameter
        color: Junction color (optional)
        uuid: Unique identifier
    """

    __token_name__ = "junction"

    at: AtXY = field(
        default_factory=lambda: AtXY(), metadata={"description": "Position"}
    )
    diameter: KiCadFloat = field(
        default_factory=lambda: KiCadFloat("diameter", 0.0),
        metadata={"description": "Junction diameter"},
    )
    color: Optional[Color] = field(
        default=None, metadata={"description": "Junction color", "required": False}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class Label(KiCadObject):
    """Local label definition token.

    The 'label' token defines a local label in the format::

        (label
            "TEXT"
            (at X Y ANGLE)
            (fields_autoplaced)
            (effects EFFECTS)
            (uuid UUID)
        )

    Args:
        text: Label text
        at: Position and rotation
        fields_autoplaced: Whether fields are autoplaced (optional)
        effects: Text effects (optional)
        uuid: Unique identifier
    """

    __token_name__ = "label"

    text: str = field(default="", metadata={"description": "Label text"})
    at: At = field(
        default_factory=lambda: At(), metadata={"description": "Position and rotation"}
    )
    fields_autoplaced: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("fields_autoplaced"),
        metadata={"description": "Whether fields are autoplaced", "required": False},
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class NoConnect(KiCadObject):
    """No connect definition token.

    The 'no_connect' token defines a no-connect symbol in the format::

        (no_connect
            (at X Y)
            (uuid UUID)
        )

    Args:
        at: Position
        uuid: Unique identifier
    """

    __token_name__ = "no_connect"

    at: AtXY = field(
        default_factory=lambda: AtXY(), metadata={"description": "Position"}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class Path(KiCadObject):
    """Path definition token.

    The 'path' token defines a hierarchical path in the format::
        (path "PATH"
            (reference "REF")
            (unit NUMBER)
        )

    Args:
        value: Path value
        reference: Component reference (optional)
        unit: Unit number (optional)
        page: Page number (optional)
    """

    __token_name__ = "path"

    value: str = field(default="", metadata={"description": "Path value"})
    reference: KiCadStr = field(
        default_factory=lambda: KiCadStr("reference", "", required=False),
        metadata={"description": "Component reference", "required": False},
    )
    unit: KiCadInt = field(
        default_factory=lambda: KiCadInt("unit", 0, required=False),
        metadata={"description": "Unit number", "required": False},
    )
    page: KiCadStr = field(
        default_factory=lambda: KiCadStr("page", "", required=False),
        metadata={"description": "Page number", "required": False},
    )


@dataclass
class Project(KiCadObject):
    """Project definition token.

    The 'project' token defines a project instance in the format::
        (project "PROJECT_NAME"
            (path "PATH"
                (reference "REF")
                (unit NUMBER)
            )
        )

    Args:
        name: Project name
        path: Hierarchical path (optional)
    """

    __token_name__ = "project"

    name: str = field(default="", metadata={"description": "Project name"})
    path: Optional[Path] = field(
        default=None,
        metadata={"description": "Hierarchical path", "required": False},
    )


@dataclass
class SheetPin(KiCadObject):
    """Sheet pin definition token.

    The 'pin' token defines a hierarchical sheet pin in the format::

        (pin "NAME" SHAPE (at X Y ANGLE) (effects EFFECTS) (uuid UUID))

    Args:
        name: Pin name string
        shape: Pin shape/direction
        at: Position and angle
        effects: Text effects (optional)
        uuid: Unique identifier
    """

    __token_name__ = "pin"

    name: str = field(default="", metadata={"description": "Pin name string"})
    shape: LabelShape = field(
        default=LabelShape.INPUT, metadata={"description": "Pin shape/direction"}
    )
    at: At = field(
        default_factory=lambda: At(), metadata={"description": "Position and angle"}
    )
    effects: Optional[Effects] = field(
        default=None, metadata={"description": "Text effects", "required": False}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class SheetLocalInstances(KiCadObject):
    """Sheet local instances definition token.

    The 'instances' token defines local sheet instances in the format::

        (instances
            (project "PROJECT_NAME"
                (path "/PATH" (page "PAGE"))
            )
        )

    Args:
        project: List of project data
    """

    __token_name__ = "instances"

    project: List[Project] = field(
        default_factory=list, metadata={"description": "List of project data"}
    )


@dataclass
class Sheet(KiCadObject):
    """Hierarchical sheet definition token.

    The 'sheet' token defines a hierarchical sheet in the format::

        (sheet
            (at X Y)
            (size WIDTH HEIGHT)
            (fields_autoplaced)
            (stroke STROKE_DEFINITION)
            (fill FILL)
            (uuid UUID)
            (property "Sheetname" "NAME")
            (property "Sheetfile" "FILE")
            (pin "NAME" SHAPE (at X Y ANGLE) (effects EFFECTS) (uuid UUID))
            ...
        )

    Args:
        at: Position
        size: Sheet size
        fields_autoplaced: Whether fields are autoplaced (optional)
        stroke: Stroke definition (optional)
        fill: Fill definition (optional)
        uuid: Unique identifier
        properties: List of properties
        pins: List of sheet pins (optional)
        exclude_from_sim: Whether sheet is excluded from simulation (optional)
        in_bom: Whether sheet appears in BOM (optional)
        on_board: Whether sheet is exported to PCB (optional)
        dnp: Do not populate flag (optional)
        rectangles: List of rectangle graphical items (optional)
        instances: Sheet local instances (optional)
    """

    __token_name__ = "sheet"

    at: AtXY = field(
        default_factory=lambda: AtXY(), metadata={"description": "Position"}
    )
    size: Size = field(
        default_factory=lambda: Size(), metadata={"description": "Sheet size"}
    )
    fields_autoplaced: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("fields_autoplaced"),
        metadata={"description": "Whether fields are autoplaced", "required": False},
    )
    stroke: Optional[Stroke] = field(
        default=None, metadata={"description": "Stroke definition", "required": False}
    )
    fill: Optional[Fill] = field(
        default=None, metadata={"description": "Fill definition", "required": False}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )
    properties: List[Property] = field(
        default_factory=list, metadata={"description": "List of properties"}
    )
    pins: Optional[List[SheetPin]] = field(
        default_factory=list,
        metadata={"description": "List of sheet pins", "required": False},
    )
    exclude_from_sim: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("exclude_from_sim"),
        metadata={
            "description": "Whether sheet is excluded from simulation",
            "required": False,
        },
    )
    in_bom: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("in_bom"),
        metadata={"description": "Whether sheet appears in BOM", "required": False},
    )
    on_board: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("on_board"),
        metadata={"description": "Whether sheet is exported to PCB", "required": False},
    )
    dnp: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("dnp"),
        metadata={"description": "Do not populate flag", "required": False},
    )
    rectangles: Optional[List[Rectangle]] = field(
        default_factory=list,
        metadata={
            "description": "List of rectangle graphical items",
            "required": False,
        },
    )
    instances: Optional[SheetLocalInstances] = field(
        default=None,
        metadata={"description": "Sheet local instances", "required": False},
    )


@dataclass
class Wire(KiCadObject):
    """Wire definition token.

    The 'wire' token defines wires in the schematic in the format::

        (wire
            COORDINATE_POINT_LIST
            STROKE_DEFINITION
            UNIQUE_IDENTIFIER
        )

    Args:
        pts: Wire connection points
        stroke: Stroke definition
        uuid: Unique identifier
    """

    __token_name__ = "wire"

    pts: Pts = field(
        default_factory=lambda: Pts(),
        metadata={"description": "Wire connection points"},
    )
    stroke: Stroke = field(
        default_factory=lambda: Stroke(), metadata={"description": "Stroke definition"}
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(), metadata={"description": "Unique identifier"}
    )


@dataclass
class SheetInstance(KiCadObject):
    """Sheet instance definition token.

    The 'path' token defines a sheet instance in the format::

        (path "PATH_STRING"
            (page "PAGE_NUMBER")
        )

    Args:
        path: Hierarchical path string
        page: Page object
    """

    __token_name__ = "path"

    path: str = field(default="", metadata={"description": "Hierarchical path string"})
    page: KiCadStr = field(
        default_factory=lambda: KiCadStr("page", ""),
        metadata={"description": "Page object"},
    )


@dataclass
class SheetInstances(KiCadObject):
    """Sheet instances container definition token.

    The 'sheet_instances' token defines sheet instances in the format::

        (sheet_instances
            (path "PATH1" (page "PAGE1"))
            (path "PATH2" (page "PAGE2"))
            ...
        )

    Args:
        sheet_instances: List of sheet instances
    """

    __token_name__ = "sheet_instances"

    sheet_instances: List[SheetInstance] = field(
        default_factory=list,
        metadata={"description": "List of sheet instances"},
    )


@dataclass
class PinRef(KiCadObject):
    """Pin reference definition token for schematic symbols.

    The 'pin' token in schematic symbols references a pin by number in the format::
        (pin "NUMBER"
            (uuid UUID)
        )

    Args:
        number: Pin number
        uuid: Unique identifier (optional)
    """

    __token_name__ = "pin"

    number: str = field(default="", metadata={"description": "Pin number"})
    uuid: Optional[Uuid] = field(
        default=None,
        metadata={"description": "Unique identifier", "required": False},
    )


@dataclass
class SymbolInstances(KiCadObject):
    """Symbol instances definition token.

    The 'instances' token defines symbol instance data in the format::
        (instances
            (project "PROJECT_NAME"
                (path "PATH"
                    (reference "REF")
                    (unit NUMBER)
                )
            )
        )

    Args:
        projects: List of project instances
    """

    __token_name__ = "instances"

    projects: List[Project] = field(
        default_factory=list,
        metadata={"description": "List of project instances"},
    )


@dataclass
class SchematicSymbol(KiCadObject):
    """Schematic symbol instance definition token.

    References a symbol definition via lib_id and adds instance-specific properties.

    The 'symbol' token in schematics defines symbol instances in the format::
        (symbol
            (lib_id "LIBRARY:SYMBOL")
            (at X Y ANGLE)
            (unit NUMBER)
            (exclude_from_sim BOOLEAN)
            (in_bom BOOLEAN)
            (on_board BOOLEAN)
            (dnp BOOLEAN)
            (fields_autoplaced BOOLEAN)
            (uuid UUID)
            PROPERTIES...
            PINS...
            (instances ...)
        )

    Args:
        lib_name: Library name (optional)
        lib_id: Library identifier referencing symbol definition (optional)
        at: Position and rotation (optional)
        mirror: Mirror transformation (optional)
        unit: Unit number (optional)
        exclude_from_sim: Whether to exclude from simulation (optional)
        in_bom: Whether to include in BOM (optional)
        on_board: Whether to place on board (optional)
        dnp: Do not populate flag (optional)
        fields_autoplaced: Whether fields are auto-placed (optional)
        uuid: Unique identifier (optional)
        properties: List of properties (optional)
        pins: List of pin references (optional)
        text: List of text elements (optional)
        instances: Symbol instances (optional)
    """

    __token_name__ = "symbol"

    lib_name: Optional[str] = field(
        default=None,
        metadata={"description": "Library name", "required": False},
    )
    lib_id: KiCadStr = field(
        default_factory=lambda: KiCadStr("lib_id", "", required=False),
        metadata={
            "description": "Library identifier referencing symbol definition",
            "required": False,
        },
    )
    at: Optional[At] = field(
        default=None,
        metadata={"description": "Position and rotation", "required": False},
    )
    mirror: KiCadStr = field(
        default_factory=lambda: KiCadStr("mirror", "x", required=False),
        metadata={"description": "Mirror transformation", "required": False},
    )
    unit: KiCadInt = field(
        default_factory=lambda: KiCadInt("unit", 0, required=False),
        metadata={"description": "Unit number", "required": False},
    )
    exclude_from_sim: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("exclude_from_sim"),
        metadata={
            "description": "Whether to exclude from simulation",
            "required": False,
        },
    )
    in_bom: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("in_bom"),
        metadata={"description": "Whether to include in BOM", "required": False},
    )
    on_board: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("on_board"),
        metadata={"description": "Whether to place on board", "required": False},
    )
    dnp: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("dnp"),
        metadata={"description": "Do not populate flag", "required": False},
    )
    fields_autoplaced: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("fields_autoplaced"),
        metadata={"description": "Whether fields are auto-placed", "required": False},
    )
    uuid: Optional[Uuid] = field(
        default=None,
        metadata={"description": "Unique identifier", "required": False},
    )
    properties: Optional[List[Property]] = field(
        default_factory=list,
        metadata={"description": "List of properties", "required": False},
    )
    pins: Optional[List[PinRef]] = field(
        default_factory=list,
        metadata={"description": "List of pin references", "required": False},
    )
    text: Optional[List[Text]] = field(
        default_factory=list,
        metadata={"description": "List of text elements", "required": False},
    )
    instances: Optional[SymbolInstances] = field(
        default=None,
        metadata={"description": "Symbol instances", "required": False},
    )


@dataclass
class KicadSch(KiCadObject):
    """KiCad schematic file definition.

    The 'kicad_sch' token defines a complete schematic file in the format::

        (kicad_sch
            (version VERSION)
            (generator GENERATOR)
            (uuid UNIQUE_IDENTIFIER)
            (lib_symbols ...)
            ;; schematic elements...
        )

    Args:
        version: File format version
        generator: Generator application name
        generator_version: Generator version (optional)
        uuid: Universally unique identifier for the schematic
        paper: Paper settings (optional)
        title_block: Title block (optional)
        lib_symbols: Symbol library container (optional)
        junctions: List of junctions (optional)
        no_connects: List of no connect markers (optional)
        bus_entries: List of bus entries (optional)
        wires: List of wires (optional)
        buses: List of buses (optional)
        labels: List of labels (optional)
        global_labels: List of global labels (optional)
        sheets: List of hierarchical sheets (optional)
        instances: List of symbol instances (optional)
        graphic_items: List of graphical items (optional)
        tables: List of tables (optional)
        hierarchical_labels: List of hierarchical labels (optional)
        rule_areas: List of rule areas (optional)
        netclass_flags: List of netclass flags (optional)
        symbols: List of symbol instances (optional)
        text: List of text elements (optional)
        images: List of image elements (optional)
        sheet_instances: Sheet instances (optional)
        embedded_fonts: Embedded fonts setting (optional)
    """

    __token_name__ = "kicad_sch"

    version: KiCadInt = field(
        default_factory=lambda: KiCadInt("version", 20240101),
        metadata={"description": "File format version"},
    )
    generator: KiCadStr = field(
        default_factory=lambda: KiCadStr("generator", ""),
        metadata={"description": "Generator application name"},
    )
    generator_version: KiCadStr = field(
        default_factory=lambda: KiCadStr("generator_version", "", required=False),
        metadata={"description": "Generator version", "required": False},
    )
    uuid: Uuid = field(
        default_factory=lambda: Uuid(),
        metadata={"description": "Universally unique identifier for the schematic"},
    )
    paper: Optional[Paper] = field(
        default=None,
        metadata={"description": "Paper settings", "required": False},
    )
    title_block: Optional[TitleBlock] = field(
        default=None,
        metadata={"description": "Title block", "required": False},
    )
    lib_symbols: Optional[LibSymbols] = field(
        default=None,
        metadata={"description": "Symbol library container", "required": False},
    )
    junctions: Optional[List[Junction]] = field(
        default_factory=list,
        metadata={"description": "List of junctions", "required": False},
    )
    no_connects: Optional[List[NoConnect]] = field(
        default_factory=list,
        metadata={"description": "List of no connect markers", "required": False},
    )
    bus_entries: Optional[List[BusEntry]] = field(
        default_factory=list,
        metadata={"description": "List of bus entries", "required": False},
    )
    wires: Optional[List[Wire]] = field(
        default_factory=list,
        metadata={"description": "List of wires", "required": False},
    )
    buses: Optional[List[Bus]] = field(
        default_factory=list,
        metadata={"description": "List of buses", "required": False},
    )
    labels: Optional[List[Label]] = field(
        default_factory=list,
        metadata={"description": "List of labels", "required": False},
    )
    global_labels: Optional[List[GlobalLabel]] = field(
        default_factory=list,
        metadata={"description": "List of global labels", "required": False},
    )
    sheets: Optional[List[Sheet]] = field(
        default_factory=list,
        metadata={"description": "List of hierarchical sheets", "required": False},
    )
    instances: Optional[List[Any]] = field(
        default_factory=list,
        metadata={"description": "List of symbol instances", "required": False},
    )
    graphic_items: Optional[
        List[Union[Arc, Bezier, Circle, Line, Polygon, Polyline, Rectangle]]
    ] = field(
        default_factory=list,
        metadata={"description": "List of graphical items", "required": False},
    )
    tables: Optional[List[Table]] = field(
        default_factory=list,
        metadata={"description": "List of tables", "required": False},
    )
    hierarchical_labels: Optional[List[HierarchicalLabel]] = field(
        default_factory=list,
        metadata={"description": "List of hierarchical labels", "required": False},
    )
    rule_areas: Optional[List[RuleArea]] = field(
        default_factory=list,
        metadata={"description": "List of rule areas", "required": False},
    )
    netclass_flags: Optional[List[NetclassFlag]] = field(
        default_factory=list,
        metadata={"description": "List of netclass flags", "required": False},
    )
    symbols: Optional[List[SchematicSymbol]] = field(
        default_factory=list,
        metadata={"description": "List of symbol instances", "required": False},
    )
    text: Optional[List[Text]] = field(
        default_factory=list,
        metadata={"description": "List of text elements", "required": False},
    )
    images: Optional[List[Image]] = field(
        default_factory=list,
        metadata={"description": "List of image elements", "required": False},
    )
    sheet_instances: Optional[SheetInstances] = field(
        default=None,
        metadata={"description": "Sheet instances", "required": False},
    )
    embedded_fonts: OptionalFlag = field(
        default_factory=lambda: OptionalFlag("embedded_fonts", token_value="yes"),
        metadata={"description": "Embedded fonts setting", "required": False},
    )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        strictness: ParseStrictness = ParseStrictness.STRICT,
        encoding: str = "utf-8",
    ) -> "KicadSch":
        """Parse from S-expression file - convenience method for schematic operations."""
        if not file_path.endswith(".kicad_sch"):
            raise ValueError("Unsupported file extension. Expected: .kicad_sch")
        with open(file_path, "r", encoding=encoding) as f:
            content = f.read()
        return cls.from_str(content, strictness)

    def save_to_file(self, file_path: str, encoding: str = "utf-8") -> None:
        """Save to .kicad_sch file format.

        Args:
            file_path: Path to write the .kicad_sch file
            encoding: File encoding (default: utf-8)
        """
        if not file_path.endswith(".kicad_sch"):
            raise ValueError("Unsupported file extension. Expected: .kicad_sch")
        content = self.to_sexpr_str()
        with open(file_path, "w", encoding=encoding) as f:
            f.write(content)



"""Project settings elements for KiCad JSON project files - .kicad_pro format."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from .base_element import ParseStrictness
from .json_base_element import JsonObject


@dataclass
class BoardDefaults(JsonObject):
    """Board design defaults settings.

    In the standard KiCad.kicad_pro file, this is an empty object {},
    so all fields are optional and represent extended configurations.
    """

    # All fields are optional since standard file has empty defaults {}
    apply_defaults_to_fp_fields: Optional[bool] = field(default=None)
    apply_defaults_to_fp_shapes: Optional[bool] = field(default=None)
    apply_defaults_to_fp_text: Optional[bool] = field(default=None)
    board_outline_line_width: Optional[float] = field(default=None)
    copper_line_width: Optional[float] = field(default=None)
    copper_text_italic: Optional[bool] = field(default=None)
    copper_text_size_h: Optional[float] = field(default=None)
    copper_text_size_v: Optional[float] = field(default=None)
    copper_text_thickness: Optional[float] = field(default=None)
    copper_text_upright: Optional[bool] = field(default=None)
    courtyard_line_width: Optional[float] = field(default=None)
    dimension_precision: Optional[int] = field(default=None)
    dimension_units: Optional[int] = field(default=None)
    dimensions: Optional[Dict[str, Any]] = field(default=None)
    fab_line_width: Optional[float] = field(default=None)
    fab_text_italic: Optional[bool] = field(default=None)
    fab_text_size_h: Optional[float] = field(default=None)
    fab_text_size_v: Optional[float] = field(default=None)
    fab_text_thickness: Optional[float] = field(default=None)
    fab_text_upright: Optional[bool] = field(default=None)
    other_line_width: Optional[float] = field(default=None)
    other_text_italic: Optional[bool] = field(default=None)
    other_text_size_h: Optional[float] = field(default=None)
    other_text_size_v: Optional[float] = field(default=None)
    other_text_thickness: Optional[float] = field(default=None)
    other_text_upright: Optional[bool] = field(default=None)
    pads: Optional[Dict[str, Any]] = field(default=None)
    silk_line_width: Optional[float] = field(default=None)
    silk_text_italic: Optional[bool] = field(default=None)
    silk_text_size_h: Optional[float] = field(default=None)
    silk_text_size_v: Optional[float] = field(default=None)
    silk_text_thickness: Optional[float] = field(default=None)
    silk_text_upright: Optional[bool] = field(default=None)
    zones: Optional[Dict[str, Any]] = field(default=None)


@dataclass
class DesignSettings(JsonObject):
    """Board design settings.

    Based on standard KiCad.kicad_pro structure.
    """

    defaults: BoardDefaults = field(default_factory=lambda: BoardDefaults())
    diff_pair_dimensions: List[Dict[str, Any]] = field(default_factory=list)
    drc_exclusions: List[Dict[str, Any]] = field(default_factory=list)
    rules: Dict[str, Any] = field(default_factory=dict)
    track_widths: List[float] = field(default_factory=list)
    via_dimensions: List[Dict[str, Any]] = field(default_factory=list)

    # Extended fields not in standard file - all optional
    meta: Optional[Dict[str, Any]] = field(default=None)
    rule_severities: Optional[Dict[str, str]] = field(default=None)
    teardrop_options: Optional[List[Dict[str, Any]]] = field(default=None)
    teardrop_parameters: Optional[List[Dict[str, Any]]] = field(default=None)
    tuning_pattern_settings: Optional[Dict[str, Any]] = field(default=None)
    zones_allow_external_fillets: Optional[bool] = field(default=None)
    zones_use_no_outline: Optional[bool] = field(default=None)


@dataclass
class IPC2581Settings(JsonObject):
    """IPC2581 export settings."""

    dist: str = field(default="")
    distpn: str = field(default="")
    internal_id: str = field(default="")
    mfg: str = field(default="")
    mpn: str = field(default="")


@dataclass
class BoardSettings(JsonObject):
    """Board-specific settings.

    Based on standard KiCad.kicad_pro structure.
    """

    design_settings: DesignSettings = field(default_factory=lambda: DesignSettings())
    ipc2581: IPC2581Settings = field(default_factory=lambda: IPC2581Settings())
    layer_pairs: List[Dict[str, Any]] = field(default_factory=list)
    layer_presets: List[Dict[str, Any]] = field(default_factory=list)
    viewports: List[Dict[str, Any]] = field(default_factory=list)
    threeD_viewports: List[Dict[str, Any]] = field(
        default_factory=list, metadata={"json_key": "3dviewports"}
    )


@dataclass
class NetClass(JsonObject):
    """Network class definition.

    Values from standard KiCad.kicad_pro file.
    """

    name: str = field(default="Default")
    bus_width: float = field(default=12.0)
    clearance: float = field(default=0.2)
    diff_pair_gap: float = field(default=0.25)
    diff_pair_via_gap: float = field(default=0.25)
    diff_pair_width: float = field(default=0.2)
    line_style: int = field(default=0)
    microvia_diameter: float = field(default=0.3)
    microvia_drill: float = field(default=0.1)
    pcb_color: str = field(default="rgba(0, 0, 0, 0.000)")
    priority: int = field(default=2147483647)
    schematic_color: str = field(default="rgba(0, 0, 0, 0.000)")
    track_width: float = field(default=0.2)
    via_diameter: float = field(default=0.6)
    via_drill: float = field(default=0.3)
    wire_width: float = field(default=6.0)


@dataclass
class NetSettings(JsonObject):
    """Network settings configuration.

    Based on standard KiCad.kicad_pro structure.
    """

    classes: List[NetClass] = field(default_factory=lambda: [NetClass()])
    meta: Dict[str, Any] = field(default_factory=lambda: {"version": 4})
    net_colors: Optional[Dict[str, Any]] = field(default=None)
    netclass_assignments: Optional[Dict[str, Any]] = field(default=None)
    netclass_patterns: List[Dict[str, Any]] = field(default_factory=list)


@dataclass
class LibrarySettings(JsonObject):
    """Library settings configuration."""

    pinned_footprint_libs: List[str] = field(default_factory=list)
    pinned_symbol_libs: List[str] = field(default_factory=list)


@dataclass
class ERCSettings(JsonObject):
    """ERC (Electrical Rules Check) settings.

    Not present in standard KiCad.kicad_pro, so all fields are optional.
    """

    erc_exclusions: Optional[List[Dict[str, Any]]] = field(default=None)
    meta: Optional[Dict[str, Any]] = field(default=None)
    pin_map: Optional[List[List[int]]] = field(default=None)
    rule_severities: Optional[Dict[str, str]] = field(default=None)


@dataclass
class SchematicBOMSettings(JsonObject):
    """BOM (Bill of Materials) settings.

    Not present in standard KiCad.kicad_pro, so all fields are optional.
    """

    annotate_start_num: Optional[int] = field(default=None)
    bom_export_filename: Optional[str] = field(default=None)
    bom_fmt_presets: Optional[List[Dict[str, Any]]] = field(default=None)
    bom_fmt_settings: Optional[Dict[str, Any]] = field(default=None)
    bom_presets: Optional[List[Dict[str, Any]]] = field(default=None)
    bom_settings: Optional[Dict[str, Any]] = field(default=None)
    connection_grid_size: Optional[float] = field(default=None)
    drawing: Optional[Dict[str, Any]] = field(default=None)
    net_format_name: Optional[str] = field(default=None)
    ngspice: Optional[Dict[str, Any]] = field(default=None)
    page_layout_descr_file: Optional[str] = field(default=None)
    plot_directory: Optional[str] = field(default=None)
    space_save_all_events: Optional[bool] = field(default=None)
    spice_adjust_passive_values: Optional[bool] = field(default=None)
    spice_current_sheet_as_root: Optional[bool] = field(default=None)
    spice_external_command: Optional[str] = field(default=None)
    spice_model_current_sheet_as_root: Optional[bool] = field(default=None)
    spice_save_all_currents: Optional[bool] = field(default=None)
    spice_save_all_dissipations: Optional[bool] = field(default=None)
    spice_save_all_voltages: Optional[bool] = field(default=None)
    subpart_first_id: Optional[int] = field(default=None)
    subpart_id_separator: Optional[int] = field(default=None)


@dataclass
class SchematicSettings(JsonObject):
    """Schematic-specific settings.

    Based on standard KiCad.kicad_pro structure.
    """

    legacy_lib_dir: str = field(default="")
    legacy_lib_list: List[str] = field(default_factory=list)

    # Extended BOM/drawing settings - not in standard file
    meta: Optional[Dict[str, Any]] = field(default=None)


@dataclass
class PcbnewSettings(JsonObject):
    """PCB editor specific settings.

    Based on standard KiCad.kicad_pro structure.
    """

    last_paths: Dict[str, str] = field(
        default_factory=lambda: {
            "gencad": "",
            "idf": "",
            "netlist": "",
            "plot": "",
            "pos_files": "",
            "specctra_dsn": "",
            "step": "",
            "svg": "",
            "vrml": "",
        }
    )
    page_layout_descr_file: str = field(default="")


@dataclass
class CvpcbSettings(JsonObject):
    """Component-Footprint assignment tool settings."""

    equivalence_files: List[str] = field(default_factory=list)


@dataclass
class ProjectMeta(JsonObject):
    """Project metadata.

    Based on standard KiCad.kicad_pro values.
    """

    filename: str = field(default="KiCad.kicad_pro")
    version: int = field(default=3)


@dataclass
class KicadProject(JsonObject):
    """KiCad project file definition (.kicad_pro format).

    The .kicad_pro file is a JSON format file that contains project configuration
    and settings for KiCad projects.

    Based on the standard KiCad.kicad_pro structure with all standard fields
    and optional extended fields for more complex projects.

    Args:
        board: Board-specific settings
        boards: List of board files in project
        cvpcb: Component-Footprint assignment settings
        libraries: Library configuration
        meta: Project metadata
        net_settings: Network classes and settings
        pcbnew: PCB editor settings
        schematic: Schematic-specific settings
        sheets: List of schematic sheets
        text_variables: Project text variables
        erc: ERC settings (optional, not in standard file)
        _original_data: Internal storage of original input data for exact round-trip
    """

    board: BoardSettings = field(default_factory=lambda: BoardSettings())
    boards: List[str] = field(default_factory=list)
    cvpcb: CvpcbSettings = field(default_factory=lambda: CvpcbSettings())
    libraries: LibrarySettings = field(default_factory=lambda: LibrarySettings())
    meta: ProjectMeta = field(
        default_factory=lambda: ProjectMeta(filename="KiCad.kicad_pro", version=3)
    )
    net_settings: NetSettings = field(default_factory=lambda: NetSettings())
    pcbnew: PcbnewSettings = field(default_factory=lambda: PcbnewSettings())
    schematic: SchematicSettings = field(default_factory=lambda: SchematicSettings())
    sheets: List[Dict[str, Any]] = field(default_factory=list)
    text_variables: Dict[str, str] = field(default_factory=dict)

    # Extended fields not in standard file - all optional
    erc: Optional[ERCSettings] = field(default=None)

    @classmethod
    def from_file(
        cls,
        file_path: str,
        strictness: ParseStrictness = ParseStrictness.STRICT,
        encoding: str = "utf-8",
    ) -> "KicadProject":
        """Parse from JSON file - convenience method for project operations.

        Args:
            file_path: Path to .kicad_pro file
            strictness: Parse strictness level (not used for JSON)
            encoding: File encoding (default: utf-8)

        Returns:
            KicadProject instance
        """
        if not file_path.endswith(".kicad_pro"):
            raise ValueError("Unsupported file extension. Expected: .kicad_pro")

        return super().from_file(file_path, strictness, encoding)

    def save_to_file(
        self, file_path: str, encoding: str = "utf-8", preserve_original: bool = False
    ) -> None:
        """Save to .kicad_pro file format.

        Args:
            file_path: Path to write the .kicad_pro file
            encoding: File encoding (default: utf-8)
            preserve_original: Whether to preserve original structure
        """
        if not file_path.endswith(".kicad_pro"):
            raise ValueError("Unsupported file extension. Expected: .kicad_pro")

        super().save_to_file(file_path, encoding, preserve_original)
